from abc import ABC, abstractmethod
from functools import wraps
from typing import Callable, Concatenate, ParamSpec, TypeVar

from .issues import PackageNotice, PackageWarning, ProductNotice, ProductWarning
from .validationdata import ValidationData

P = ParamSpec('P')
T = TypeVar('T')
R = TypeVar('R', bound='Ruleset')

def rule(func: Callable[Concatenate[R, P], T]) -> Callable[Concatenate[R, P], T]:
	"""Decorator to print rule name."""

	@wraps(func)
	def wrapper(self: R, *args: P.args, **kwargs: P.kwargs) -> T:
		if self.data.verbose:
			if len(args):
				print(func.__qualname__, [*args])
			else:
				print(func.__qualname__)
		return func(self, *args, **kwargs)

	return wrapper

class Ruleset(ABC):
	"""Abstract base class for validation rules.

	Arguments:
		data (ValidationData): validation data of product to validate.
	"""

	def __init__(self, data: ValidationData) -> None:
		self.data = data
		self._validate()

	@abstractmethod
	def _validate(self) -> None:
		if self.data.verbose:
			print (self.__class__.__qualname__)

class PackageRuleset(Ruleset):
	"""Abstract base class for package validation rules.

	Arguments:
		data (ValidationData): validation data of product to validate.
	"""

	def _addIssue(self, issue: PackageNotice | PackageWarning) -> None:
		"""Add issue to validation data."""

		self.data.issues.package.append(issue)

class ProductRuleset(Ruleset):
	"""Abstract base class for product validation rules.

	Arguments:
		data (ValidationData): validation data of product to validate.
	"""

	def _addIssue(self, issue: ProductNotice | ProductWarning) -> None:
		"""Add issue to validation data."""

		self.data.issues.product.append(issue)


import gzip
import re

from contextlib import contextmanager
from difflib import SequenceMatcher
from mimetypes import add_type, guess_all_extensions
from pathlib import Path
from typing import cast, BinaryIO, Generator, NamedTuple

from fs.base import FS
from fs.errors import InvalidCharsInPath
from fs.info import Info
from fs.path import basename, combine, splitext
from PIL import Image

from .validationdata import ValidationData

add_type('image/x-tga', '.tga')

class Step(NamedTuple):
	"""Typing information for pyFilesystem2 Step"""

	path: str
	dirs: list[Info]
	files: list[Info]

alpha_numeric_regex = re.compile(r'[^0-9A-Za-z]')
def alpha_numeric(string: str) -> str:
	"""Remove all non alphanumeric characters from a string.

	Arguments:
		string (str): String to process.
	"""

	return alpha_numeric_regex.sub('', string)

def thumbnailsFor(fs: FS, path: str) -> tuple[list[str], list[str]]:
	"""Get list of thumbnails (and tip images) for a file.

	Returns a tuple of existing files in
		[filename.png, basefilename.png]
		[filename.tip.png, basefilename.tip.png]

	Arguments:
		fs (fs.FS): PyFilesystem2 filesystem to use.
		path (str): File to get thumbnails from.
	"""

	thumbnails: list[str] = []
	tips: list[str] = []

	if fs.exists(file := path + '.png'):
		thumbnails.append(file)
	if fs.exists(file := splitext(path)[0] + '.png'):
		thumbnails.append(file)

	if fs.exists(file := path + '.tip.png'):
		tips.append(file)
	if fs.exists(file := splitext(path)[0] + '.tip.png'):
		tips.append(file)

	return (thumbnails, tips)

def checkDirectoryHasSelfAsChild(fs: FS, dir: str) -> bool:
	"""Check directory does not contain subdirectory with same name.

	Arguments:
		data (ValidationData): validation data of product to validate.
		dir (str): Root directory to check
	"""

	return fs.isdir(dir) and fs.isdir(combine(dir, basename(dir)))

def checkVendorDirsOnly(data: ValidationData, dir: str) -> list[str]:
	"""Check directory contains only subdirectories.

	Adds subdirectories to data.vendor_paths and returns a list of files in directory.

	Arguments:
		data (ValidationData): validation data of product to validate.
		dir (str): Root directory to check
	"""

	# Get case sensitive version of path from filesystem
	_, del_path = data.product_fs.delegate_path(dir)

	root_files: list[str] = []

	if data.product_fs.isdir(del_path):
		for entry in data.product_fs.scandir(del_path):
			if entry.is_file:
				root_files.append(entry.name)
			else:
				data.vendor_paths.setdefault(entry.name, set()).add(entry.make_path(del_path))

	return root_files

def checkImageDir(fs: FS, dir: str, preferred_suffixes: set[str]) -> tuple[list[str], list[str], list[str], dict[str, str]]:
	"""Check Image directory only contains supported images.

	Returns a tuple of (
		non image files: list[str],
		supported image files without preferred suffixes: list[str],
		unreadable image files (for filetypes supported by both DS and `PIL`): list[str],
		image files with incorrect file extensions and their detected MIME type: list[tuple[str, str]]
	)

	Arguments:
		fs (fs.FS): PyFilesystem2 filesystem to use.
		dir (str): Root directory to check.
		preferred_suffixes (set[str]): List of preferred file extensions.
	"""

	DS_SUPPORTED_TEXTURE_SUFFIXES = set(['.bmp', '.bum', '.gif', '.jpeg', '.jpg', '.png', '.hdr', '.exr', '.tif', '.tiff', '.tga'])
	PIL_UNSUPPORTED_IMAGE_SUFFIXES = set(['.dsi', '.svg', '.hdr', '.exr', '.bum'])

	non_image_files: list[str] = []
	atypical_image_files: list[str] = []
	unreadable_image_files: list[str] = []
	incorrect_image_suffixes: dict[str,str] = {}

	if fs.isdir(dir):
		for file in fs.walk.files(dir):
			path = Path(file)
			suffix = path.suffix.lower()
			if suffix not in preferred_suffixes.union(DS_SUPPORTED_TEXTURE_SUFFIXES):
				non_image_files.append(file)
			else:
				if suffix not in preferred_suffixes:
					atypical_image_files.append(file)
				try:
					with Image.open(fs.openbin(file)) as im:
						im.verify()
						if (mime_type := cast(str | None, im.get_format_mimetype())) and suffix not in guess_all_extensions(mime_type):  # pyright: ignore[reportGeneralTypeIssues, reportUnknownMemberType]
							incorrect_image_suffixes[file] = mime_type
				except Exception:
					if suffix not in PIL_UNSUPPORTED_IMAGE_SUFFIXES:
						unreadable_image_files.append(file)

	return (non_image_files, atypical_image_files, unreadable_image_files, incorrect_image_suffixes)

def trackDependencyIfExists(data: ValidationData, file: str, used_by_file: str) -> bool:
	"""Check if file exists (case insensitive) and track dependency.

	Adds file to data.dependency_files and returns true if file exists.

	Arguments:
		data (ValidationData): validation data of product to validate.
		file (str): File path to check existence of.
		used_by_file (str): File path to record as cause of dependency.
	"""

	exists = True

	try:
		# Get case sensitive version of path from filesystem
		_, del_path = data.filesystem.delegate_path(file)

		if not data.filesystem.exists(del_path):
			exists = False
		elif data.product_fs.exists(del_path):
			data.referenced_files.setdefault(del_path, set()).add(used_by_file)
		else:
			fs_name, _ = data.filesystem_unwrapped.which(del_path)
			fs_name = cast(str, fs_name)
			data.dependency_files.setdefault(fs_name, {}).setdefault(del_path, set()).add(used_by_file)
	except InvalidCharsInPath:
		exists = False

	return exists

def checkTypo(string: str, known_strings: list[str], ratio: float = 0.8) -> list[str]:
	"""Check string is in list or a near match for element in list.

	Returns a list of matches or near matches.

	Arguments:
		string (str): String to check.
		known_strings (list[str]): List of strings to check against.
		ratio (float): Minimum `SequenceMatcher` ratio to match.
	"""

	s = SequenceMatcher(b = string)	# b parameter is cached
	matches = [e for e in known_strings if s.set_seq1(e) or s.ratio() > ratio]

	return matches

@contextmanager
def decompressDSON(file: BinaryIO) -> Generator[BinaryIO, None, None]:
	"""Decompress DSON file if necessary.

	Returns a BinaryIO.

	Arguments:
		file (BinaryIO): File to decompress
	"""
	try:
		g = gzip.open(file, 'r')
		g.peek(1)
		file = cast(BinaryIO, g)
	except gzip.BadGzipFile:
		file.seek(0)

	yield file


from pathlib import Path

from PIL import Image

from ..issues import support as issues
from ..validator.ruleset import ProductRuleset, rule

_SUPPORT_DIR = 'Runtime/Support'

class ValidateRuntimeSupportDirectory(ProductRuleset):
	"""Perform Runtime/Support validation of product.

	Arguments:
		data (ValidationData): validation data of product to validate.
	"""

	def _validate(self) -> None:
		"""Perform validation."""

		super()._validate()
		if self.data.product_fs.isdir(_SUPPORT_DIR):
			self._checkDirectoriesInSupportDirectory()
			self._checkMetadataFiles()
		else:
			self._addIssue(issues.MissingMetadataFilesIssue())

	@rule
	def _checkDirectoriesInSupportDirectory(self) -> None:
		"""Check product for directories in Runtime/Support directory."""

		if subdirectories := [entry.name for entry in self.data.product_fs.scandir(_SUPPORT_DIR) if entry.is_dir]:
			self._addIssue(issues.SubdirectoriesInRuntimeSupportIssue(subdirectories))

	@rule
	def _checkMetadataFiles(self) -> None:
		"""Check metadata files in Runtime/Support directory."""

		IMAGE_SUFFIXES = set(['.jpg', '.png'])

		# User guide states icon should be 114x148 (or 100×143 for old DAZ products),
		# but reference guide states icon may be larger and will be automatically
		# scaled as needed.
		IMAGE_MINIMUM_DIMENSIONS = (114, 148)

		metadata: dict[str, list[str]] = {}
		for file in [Path(f) for f in self.data.product_fs.walk.files(_SUPPORT_DIR)]:  # pyright: ignore[reportUnknownMemberType]
			if not file.stem in metadata:
				metadata[file.stem] = []
			metadata[file.stem].append(file.suffix)

		dsx_files: list[str] = []
		unexpected_files: list[str] = []
		missing_icon_files: list[str] = []
		missing_script_files: list[str] = []
		redundant_icon_files: list[str] = []
		undersized_icon_files: dict[str, tuple[int, int]] = {}
		supportDirPath = Path(_SUPPORT_DIR)
		for entry in metadata:
			if not '.dsx' in metadata[entry]:
				for suffix in metadata[entry]:
					unexpected_files.append((supportDirPath / f"{entry}{suffix}").as_posix())
			else:
				dsx_files.append((supportDirPath / f"{entry}.dsx").as_posix())
				if not '.dsa' in metadata[entry]:
					missing_script_files.append((supportDirPath / f"{entry}.dsa").as_posix())

				if not any (suffix in IMAGE_SUFFIXES for suffix in metadata[entry]):
					missing_icon_files.append((supportDirPath / f"{entry}.({'|'.join([s.lstrip('.') for s in IMAGE_SUFFIXES])})").as_posix())
				else:
					image_suffixes = IMAGE_SUFFIXES.intersection(metadata[entry])
					if len(image_suffixes) > 1:
						for suffix in image_suffixes:
							redundant_icon_files.append((supportDirPath / f"{entry}{suffix}").as_posix())

					for suffix in image_suffixes:
						with self.data.product_fs.openbin((supportDirPath / f"{entry}{suffix}").as_posix()) as file:
							dimensions = Image.open(file).size
							if any (i < j for i, j in zip(dimensions, IMAGE_MINIMUM_DIMENSIONS)):
								undersized_icon_files[(supportDirPath / f"{entry}{suffix}").as_posix()] = dimensions

				if unexpected_suffixes := list(set(metadata[entry]) - set(['.dsa', '.dsx']) - IMAGE_SUFFIXES):
					for suffix in unexpected_suffixes:
						unexpected_files.append((supportDirPath / f"{entry}{suffix}").as_posix())

		if not dsx_files:
			self._addIssue(issues.MissingMetadataFilesIssue())

		if unexpected_files:
			self._addIssue(issues.UnexpectedFilesInRuntimeSupportIssue(unexpected_files))

		if missing_icon_files:
			self._addIssue(issues.MissingMetadataIconFilesIssue(missing_icon_files))

		if missing_script_files:
			self._addIssue(issues.MissingMetadataScriptFilesIssue(missing_script_files))

		if redundant_icon_files:
			self._addIssue(issues.RedundantMetadataIconFilesIssue(redundant_icon_files))

		if undersized_icon_files:
			self._addIssue(issues.UndersizedMetadataIconFilesIssue(undersized_icon_files))


import sys

from argparse import ArgumentParser, ArgumentTypeError, FileType
from fs.errors import ResourceNotFound
from fs.opener import open_fs
from fs.opener.parse import parse_fs_url
from fs.path import relpath
from jinja2 import Environment, FileSystemLoader, PackageLoader, Template, select_autoescape
from jinja2.exceptions import TemplateError
from pathlib import Path
from pprint import PrettyPrinter
from typing import Any

from .validator import ValidationData, validate

class PrettyPrinterWithoutStringWrapping(PrettyPrinter):
	"""PrettyPrinter without string wrapping."""

	def __init__(self, *args: Any) -> None:
		super().__init__(*args)
		self._dispatch[str.__repr__] = self._pprint_str  # pyright: ignore[reportGeneralTypeIssues, reportUnknownMemberType]

	def _pprint_str(self, _self: PrettyPrinter, *args: Any) -> None:
		width = self._width
		self._width = sys.maxsize
		try:
			super()._pprint_str(*args)  # pyright: ignore[reportGeneralTypeIssues, reportUnknownMemberType]
		finally:
			self._width = width

def _content_location(path: str) -> Path:
	"""Clean up error messages for paths passed on commandline."""

	try:
		return Path(path).resolve(True)
	except FileNotFoundError as e:
		url = f'zip://{path}'
		try:
			parsed_url = parse_fs_url(url)
			fs = open_fs(url)
			fs.opendir(relpath((parsed_url.path or '').lstrip('\\')))  # pyright: ignore[reportUnknownArgumentType, reportUnknownMemberType]
		except ResourceNotFound as e:
			raise ArgumentTypeError(f'{parsed_url.path} not found in {parsed_url.resource}')  # pyright: ignore[reportUnboundVariable, reportUnknownMemberType]
		except:
			raise ArgumentTypeError(e)
		else:
			return Path(path)
	except Exception as e:
		raise ArgumentTypeError(e)

def _jinja_user_template(template_filepath: str) -> Template:
	"""Load user specified Jinja template."""

	filepath = Path(template_filepath)
	if not filepath.exists():
		raise ArgumentTypeError(f"Template not found: {template_filepath}")

	jinja_env = Environment(loader=FileSystemLoader(searchpath=filepath.parent), auto_reload=False)
	jinja_env.filters['pformat'] = PrettyPrinterWithoutStringWrapping().pformat  # pyright: ignore[reportGeneralTypeIssues, reportUnknownMemberType]

	try:
		return jinja_env.get_template(filepath.name)
	except TemplateError as e:
		raise ArgumentTypeError(repr(e))
	except Exception as e:
		raise ArgumentTypeError(e)

def _main() -> None:
	parser = ArgumentParser(description = 'Validate DAZ Studio Product', epilog = 'For non-DIM ZIP files, a subdirectory may be specified as the content root by appending !<subdirectory> to the filename: "example.zip!My Library"')
	parser.add_argument('product_path', type = _content_location, nargs = '+', help = 'DIM Packages, ZIP files, or Content Directories to validate')
	parser.add_argument('-d', '--dependencies', type = _content_location, nargs = '+', help = 'additional DIM Packages, ZIP files, or Content Directories which are not validated but which the validated product depends on')

	parser.add_argument('-D', '--daz', action = 'store_true', help = 'enable validation rules for products distributed by Daz Productions, Inc')
	parser.add_argument('-O', '--daz-original', action = 'store_true', help = 'enable validation rules for products produced by Daz Productions, Inc')

	parser.add_argument('-p', '--poser', action = 'store_true', help = 'enable validation rules for included Poser content (experimental)')
	parser.add_argument('-s', '--dson-schema', action = 'store_true', help = 'enable DSON schema checking (experimental)')

	template_parser = parser.add_mutually_exclusive_group()
	template_parser.add_argument('-H', '--html', action = 'store_true', help = 'generate HTML report')
	template_parser.add_argument('-T', '--template', type = _jinja_user_template, help = 'generate report using Jinja template')

	parser.add_argument('-o', '--output', type = FileType('w', encoding='utf-8'), default=sys.stdout, help = 'write report to file')
	parser.add_argument('-v', '--verbose', action = 'store_true', help = 'enable verbose output')

	args = parser.parse_args()

	# Prepare report template
	if args.template:
		template = args.template
	else:
		jinja_env = Environment(loader=PackageLoader(__package__, "data/templates",), autoescape=select_autoescape(['html.jinja']), auto_reload=False)
		jinja_env.filters['pformat'] = PrettyPrinterWithoutStringWrapping().pformat  # pyright: ignore[reportGeneralTypeIssues, reportUnknownMemberType]
		template = jinja_env.get_template("report.html.jinja" if args.html else "report.txt.jinja")

	# Validate
	data = ValidationData(args.product_path, args.dependencies, args.daz, args.daz_original, args.poser, args.dson_schema, args.verbose)
	validate(data)

	args.output.write(template.render(data=data))

def _profile() -> None:  # pyright: ignore[reportUnusedFunction]
	import cProfile, pstats
	profiler = cProfile.Profile()
	profiler.enable()
	_main()
	profiler.disable()
	pstats.Stats(profiler).dump_stats('stats.profiler')
 
if __name__ == '__main__':
	_profile()
	# _main()


from pathlib import Path
from typing import cast, Type, TypeVar

from fs.base import FS
from fs.dimzipfs import DIMZipFS
from fs.errors import CreateFailed
from fs.multifs import MultiFS
from fs.opener import open_fs
from fs.opener.errors import OpenerError
from fs.opener.parse import parse_fs_url
from fs.path import relpath
from fs.wrapcifs import WrapCaseInsensitive
from fs.zipfs import ReadZipFS

from .issues import PackageNotice, PackageWarning, ProductNotice, ProductWarning
from .package import Package

class ValidationData:
	"""Stores data on what to validate and data discovered during validation.

	Arguments:
		product_path (list[pathlib.Path]): A list of files or directories making up the product to validate
		dependencies_paths (list[pathlib.Path]): A list of files or directories the product may reference
		                                         but that are not part of the product to validate
		daz (bool): Validate as if this is a product distributed by Daz Productions, Inc
		daz_original (bool): Validate as if this is a product produced by Daz Productions, Inc
		poser (bool): Validate using rules for Poser content (experimental)
		dson_schema (bool): Validate DSON schema (experimental)
		verbose (bool): Enable verbose output
	"""

	class Issues:
		"""Stores issues discovered during validaiton."""

		def __init__(self) -> None:
			self.package: list[PackageNotice | PackageWarning] = []
			self.product: list[ProductNotice | ProductWarning] = []

	class MetaData:
		"""Stores metadata information discovered during validation."""

		def __init__(self) -> None:
			self.products: list[str] = []
			self.artists: set[str] = set()
			self.stores: set[str] = set()

	def __init__(self, product_paths: list[Path], dependencies_paths: list[Path] | None = None, daz: bool = False, daz_original: bool = False, poser: bool = False, dson_schema: bool = False, verbose: bool = False) -> None:
		self.product_paths = product_paths
		self.dependency_paths = dependencies_paths
		self.poser = poser
		self.daz = daz or daz_original
		self.daz_original = daz_original
		self.dson_schema = dson_schema
		self.verbose = verbose

		self.product_fs_unwrapped = _create_merged_fs(product_paths)
		self.product_fs = WrapCaseInsensitive(self.product_fs_unwrapped)
		self.filesystem_unwrapped = _create_merged_fs((dependencies_paths or []) + product_paths)
		self.filesystem = WrapCaseInsensitive(self.filesystem_unwrapped)
		
		self.packages = [Package(dfs) for fs in self.product_fs_unwrapped.iterate_fs() if (dfs := _get_unwrapped_fs(fs[1], DIMZipFS))]
		self.zips = [zfs for fs in self.product_fs_unwrapped.iterate_fs() if (zfs := _get_unwrapped_fs(fs[1], ReadZipFS))]

		self.issues = self.Issues()
		self.metadata = self.MetaData()
		self.contributors: dict[str, dict[str, set[str]]] = {}

		self.vendor_paths: dict[str, set[str]] = {}
		self.dependency_files: dict[str, dict[str, set[str]]] = {}
		self.referenced_files: dict[str, set[str]] = {}
		self.missing_referenced_files: dict[str, set[str]] = {}
		self.shader_users: dict[str, set[str]] = {}
		self.postload_files: set[str] = set()


def _create_merged_fs(filesystems: list[Path]) -> MultiFS:
	def _path_to_fs(path: Path) -> FS:
		for f in [f'dimzip://{path}', f'zip://{path}', f'osfs://{path}']:
			try:
				if isinstance(fs := open_fs(f), DIMZipFS):
					if fs.isdir('Content'):
						fs = fs.opendir('Content')  # pyright: ignore[reportUnknownMemberType]
					else:
						raise ValueError(f'DIM ZIP does not have Content: {path}')
				elif isinstance(fs, ReadZipFS):
					if subpath := cast(str | None, parse_fs_url(f).path):  # pyright: ignore[reportUnknownMemberType]
						fs = fs.opendir(relpath(subpath.lstrip('\\')))  # pyright: ignore[reportUnknownMemberType]
				return fs
			except (CreateFailed, OpenerError):
				continue

		raise ValueError(f'Not a ZIP file or directory: {path}')

	fs = MultiFS()
	for f in filesystems:
		fs.add_fs(f.as_posix(), _path_to_fs(f))

	return fs

T = TypeVar('T', bound=FS)
def _get_unwrapped_fs(fs: FS, fs_type: Type[T]) -> T | None:
	if isinstance(fs, fs_type):
		return fs
	else:
		try:
			return _get_unwrapped_fs(fs.delegate_fs(), fs_type)  # pyright: ignore[reportGeneralTypeIssues, reportUnknownArgumentType, reportUnknownMemberType]
		except AttributeError:
			pass


import zlib

from fs.path import combine, split

from ..issues import geometries as issues
from ..validator.ruleset import ProductRuleset, rule
from ..validator.utilities import checkDirectoryHasSelfAsChild, checkVendorDirsOnly

_GEOMETRIES_DIR = 'Runtime/Geometries'

class ValidateRuntimeGeometriesDirectory(ProductRuleset):
	"""Perform Runtime/Geometries validation of product.

	Arguments:
		data (ValidationData): validation data of product to validate.
	"""

	def _validate(self) -> None:
		"""Perform validation."""

		super()._validate()
		if self.data.product_fs.isdir(_GEOMETRIES_DIR):
			self._checkExcessGeometriesDirectory()
			self._checkVendorDirsOnlyInGeometriesDirectory()
			self._checkValidCompressedGeometryFiles()
			self._checkGeometriesFiles()

			# TODO: Validate .obj and .mtl files

	@rule
	def _checkExcessGeometriesDirectory(self) -> None:
		"""Check product for Textures directory in Runtime/Geometries directory."""

		if checkDirectoryHasSelfAsChild(self.data.product_fs, _GEOMETRIES_DIR):
			self._addIssue(issues.ExcessGeometriesDirectoryIssue([combine(_GEOMETRIES_DIR, split(_GEOMETRIES_DIR)[1])]))

	@rule
	def _checkVendorDirsOnlyInGeometriesDirectory(self) -> None:
		"""Check product has only Vendor directories in Runtime/Geometries.

		Updates `data.vendor_paths` with found vendor directories.
		"""

		if geometries_root_files := checkVendorDirsOnly(self.data, _GEOMETRIES_DIR):
			self._addIssue(issues.FilesInRootOfGeometriesDirectoryIssue(geometries_root_files))

	@rule
	def _checkValidCompressedGeometryFiles(self) -> None:
		"""Check compressed files in Runtime/Geometries directory are valid zlib compressed files."""

		invalid_zlib_files: list[str] = []
		for file in self.data.product_fs.walk.files(_GEOMETRIES_DIR, filter=['*.obz']):  # pyright: ignore[reportUnknownMemberType]
			try:
				# NOTE: Poser compressed files should be zlib compressed, but DAZ Studio can read gzip as well, so allow it.
				zlib.decompress(self.data.product_fs.readbytes(file), zlib.MAX_WBITS|32)
			except zlib.error:
				invalid_zlib_files.append(file)

		if invalid_zlib_files:
			self._addIssue(issues.InvalidCompressedFilesInRuntimeGeometriesDirectoryIssue(invalid_zlib_files))

	@rule
	def _checkGeometriesFiles(self) -> None:
		"""Check files in Runtime/Geometries."""

		if unexpected_files := [file for file in self.data.product_fs.walk.files(_GEOMETRIES_DIR, exclude=['*.obj', '*.obz', '*.mtl'])]:  # pyright: ignore[reportUnknownMemberType]
			self._addIssue(issues.UnexpectedFilesInRuntimeGeometriesDirectoryIssue(unexpected_files))


from ..issues import manifest as issues
from ..validator.package import Package
from ..validator.ruleset import PackageRuleset, rule
from ..validator.schema import SchemaCheckedXML

_DSX_FILES = ['Manifest.dsx', 'Supplement.dsx']

class ValidatePackageManifests(PackageRuleset):
	"""Perform manifest validation of package.

	Arguments:
		data (ValidationData): validation data of package to validate.
	"""

	def _validate(self) -> None:
		"""Perform validation."""

		super()._validate()

		invalid_manifest_files: dict[str, Exception] = {}
		self.missing_manifest_files: dict[str, list[str]] = {}
		self.unlisted_manifest_files: dict[str, list[str]] = {}
		self.extraneous_root_files: dict[str, list[str]] = {}
		for package in self.data.packages:
			try:
				manifest = SchemaCheckedXML(package.root_fs.openbin('Manifest.dsx'), 'Manifest.xsd').tree
			except Exception as e:
				invalid_manifest_files[package.path.as_posix()] = e
			else:
				self.manifest_files = [e.attrib['VALUE'] for e in manifest.xpath("/DAZInstallManifest/File[@ACTION='Install']")]

				self._checkMissingManifestFiles(package)
				self._checkUnlistedManifestFiles(package)
				self._checkRootExtraneousFiles(package)

		if invalid_manifest_files:
			self._addIssue(issues.InvalidPackageManifestIssue(invalid_manifest_files))

		if self.missing_manifest_files:
			self._addIssue(issues.MissingPackageManifestFilesIssue(self.missing_manifest_files))

		if self.unlisted_manifest_files:
			self._addIssue(issues.MissingPackageManifestRecordIssue(self.unlisted_manifest_files))

		if self.extraneous_root_files:
			self._addIssue(issues.RootFilesInPackageIssue(self.extraneous_root_files))

	@rule
	def _checkMissingManifestFiles(self, package: Package) -> None:
		"""Check for files listed in Manifest but not included in zip file."""

		if missing_files := [f for f in self.manifest_files if not package.root_fs.exists(f)]:
			self.missing_manifest_files[package.path.name] = missing_files

	@rule
	def _checkUnlistedManifestFiles(self, package: Package) -> None:
		"""Check for files not listed in Manifest but included in zip file."""

		if extra_files := [file for file in (file.lstrip('/') for file in package.root_fs.walk.files()) if not file in self.manifest_files + _DSX_FILES]:
			self.unlisted_manifest_files[package.path.name] = extra_files

	@rule
	def _checkRootExtraneousFiles(self, package: Package) -> None:
		"""Check for unexpected files in root of zip file."""

		if root_files := [file for file in (entry.name.lstrip('/') for entry in package.root_fs.scandir('') if entry.is_file and not entry.name in _DSX_FILES)]:
			self.extraneous_root_files[package.path.name] = root_files


from . import PackageWarning

class PackageNameIssue(PackageWarning):
	title = 'Invalid Package Filename'
	reference = 'http://docs.daz3d.com/doku.php/public/software/install_manager/referenceguide/tech_articles/package_naming/'
	description = (
		"Package filenames must match the regular expression "
		"'^([A-Z][0-9A-Z]{0,6})(?=\\d{8})(\\d{8})(-(\\d{2}))?_([0-9A-Za-z]+)\\.zip$'")

class CustomPackageNameIssue(PackageWarning):
	title = 'Custom Package Filename'
	reference = 'http://docs.daz3d.com/doku.php/public/software/install_manager/referenceguide/tech_articles/package_naming/'
	description = "Packages distributed by Daz Productions, Inc should have a source prefix of 'IM', 'DZ', 'DAZ', 'DAZ3D', or 'TAFI'"

class StandardPackageNameIssue(PackageWarning):
	title = 'Standard Package Filename'
	reference = 'http://docs.daz3d.com/doku.php/public/software/install_manager/referenceguide/tech_articles/package_naming/'
	description = "Packages not distributed by Daz Productions, Inc must not have a source prefix of 'IM', 'DZ', 'DAZ', 'DAZ3D', or 'TAFI'"


from fs.path import combine, join, iteratepath

from ..issues import data as issues
from ..validator.ruleset import ProductRuleset, rule
from ..validator.utilities import checkDirectoryHasSelfAsChild

_DATA_DIR = 'data'
_ITEM_SUBDIRECTORIES_WITH_VENDOR_SUBDIRECTORIES = ['add-ons', 'morphs', 'uv sets', 'projection morphs']
_ITEM_SUBDIRECTORIES_TO_SKIP_REFERENCE_CHECK = ['projection templates', 'tools']

# http://docs.daz3d.com/doku.php/public/software/dazstudio/4/referenceguide/tech_articles/file_extensions/start#serialized_binary
_LEGACY_FILE_EXTENSIONS = ['*.dsd', '*.dso', '*.dsv']


class ValidateDataDirectory(ProductRuleset):
	"""Perform data validation of product.

	Arguments:
		data (ValidationData): validation data of product to validate.
	"""

	def _validate(self) -> None:
		"""Perform validation."""

		super()._validate()
		if self.data.product_fs.isdir(_DATA_DIR):
			self._checkExcessDataDirectory()
			self._checkAutoAdaptedDirectory()
			self._checkLegacyDirectories()
			self._checkVendorDirectories()
			self._checkRootFiles()
			self._checkLegacyFormatFiles()
			self._checkUserFacingFiles()
			self._checkVendorDirectoriesHaveFiles()
			self._checkVendorProductDirectoriesHaveFiles()
			self._checkUnreferencedFiles()

	@rule
	def _checkExcessDataDirectory(self) -> None:
		"""Check product for data directory in data directory."""

		if checkDirectoryHasSelfAsChild(self.data.product_fs, _DATA_DIR):
			self._addIssue(issues.ExcessDataDirectoryIssue([combine(_DATA_DIR, _DATA_DIR)]))

	@rule
	def _checkAutoAdaptedDirectory(self) -> None:
		"""Check product for auto_adapted in data directory."""

		AUTO_DIR = combine(_DATA_DIR, 'auto_adapted')
		if self.data.product_fs.isdir(AUTO_DIR):
			self._addIssue(issues.AutoAdaptedInDataDirectoryIssue([AUTO_DIR]))

	@rule
	def _checkLegacyDirectories(self) -> None:
		"""Check product for legacy directories in data directory."""

		LEGACY_DIRECTORIES = ['3_0', '4_0', '4_0_2']

		if legacy_directories := [dir for dir in LEGACY_DIRECTORIES if self.data.product_fs.isdir(combine(_DATA_DIR, dir))]:
			self._addIssue(issues.LegacyDirectoriesInDataDirectoryIssue(legacy_directories))

	@rule
	def _checkRootFiles(self) -> None:
		"""Check product for files in root data directory."""

		if root_files := [entry.name for entry in self.data.product_fs.scandir(_DATA_DIR) if entry.is_file]:
			self._addIssue(issues.FilesInRootOfDataDirectoryIssue(root_files))

	@rule
	def _checkVendorDirectories(self) -> None:
		"""Check product for data/Vendor, data/Vendor/Product/Item/*/Vendor directories.

		Updates `data.vendor_paths` with found vendor directories.
		"""

		# NOTE: This is guesswork as the directory structure is not required, just common

		for path, info in self.data.product_fs.walk.info(_DATA_DIR):  # pyright: ignore[reportUnknownMemberType]
			if info.is_file:
				path_parts = iteratepath(path)
				# data / Vendor / Product / Item / ? / Vendor / Base / File
				if len(path_parts) > 6 and path_parts[4].lower() in _ITEM_SUBDIRECTORIES_WITH_VENDOR_SUBDIRECTORIES:
					self.data.vendor_paths.setdefault(path_parts[5], set()).add(join(*path_parts[:6]))
				elif len(path_parts) > 2:
					self.data.vendor_paths.setdefault(path_parts[1], set()).add(combine(path_parts[0], path_parts[1]))

	@rule
	def _checkVendorDirectoriesHaveFiles(self) -> None:
		"""Check product for files in data/Vendor directories."""

		vendor_root_files: list[str] = []
		for vendor in self.data.product_fs.scandir(_DATA_DIR):
			if vendor.is_dir:
				vendor_root_files.extend([entry.make_path(vendor.name) for entry in self.data.product_fs.scandir(vendor.make_path(_DATA_DIR)) if entry.is_file])

		if vendor_root_files:
			self._addIssue(issues.FilesInDataVendorDirectoryIssue(vendor_root_files))

	@rule
	def _checkVendorProductDirectoriesHaveFiles(self) -> None:
		"""Check product for files in data/Vendor/Product directories."""

		product_root_files: list[str] = []
		for vendor in self.data.product_fs.scandir(_DATA_DIR):
			if vendor.is_dir:
				vendorPath = vendor.make_path(_DATA_DIR)
				for product in self.data.product_fs.scandir(vendorPath):
					product_root_files = [entry.make_path(combine(vendor.name, product.name)) for entry in self.data.product_fs.scandir(vendorPath) if entry.is_file]

		if product_root_files:
			self._addIssue(issues.FilesInDataProductDirectoryIssue(product_root_files))

	@rule
	def _checkLegacyFormatFiles(self) -> None:
		"""Check product for legacy format files in data directory."""

		if legacy_format_files := [entry.lstrip('/') for entry in self.data.product_fs.walk.files(_DATA_DIR, filter=_LEGACY_FILE_EXTENSIONS)]:  # pyright: ignore[reportUnknownMemberType]
			self._addIssue(issues.LegacyFilesInDataDirectoryIssue(legacy_format_files))

	@rule
	def _checkUserFacingFiles(self) -> None:
		"""Check product for user facing files in data directory."""

		if duf_files := [entry.lstrip('/') for entry in self.data.product_fs.walk.files(_DATA_DIR, filter=['*.duf']) if entry not in self.data.postload_files]:  # pyright: ignore[reportUnknownMemberType]
			self._addIssue(issues.DufFilesInDataDirectoryIssue(duf_files))

	@rule
	def _checkUnreferencedFiles(self) -> None:
		"""Check product for unreferenced files in data directory."""

		# TODO: Filter out _ITEM_SUBDIRECTORIES_WITH_VENDOR_SUBDIRECTORIES files in a better manner for other products?

		referenced_files_lc = {f.lower() for f in self.data.referenced_files}

		unreferenced_files: list[str] = []
		for file in self.data.product_fs.walk.files(_DATA_DIR, exclude=_LEGACY_FILE_EXTENSIONS):  # pyright: ignore[reportUnknownMemberType]
			if (not '/' + file.lower() in referenced_files_lc
				and not (len(filepath := iteratepath(file)) > 4 and filepath[4].lower() in _ITEM_SUBDIRECTORIES_WITH_VENDOR_SUBDIRECTORIES + _ITEM_SUBDIRECTORIES_TO_SKIP_REFERENCE_CHECK)):
				unreferenced_files.append(file)

		if unreferenced_files:
			self._addIssue(issues.UnreferencedFilesInDataDirectoryIssue(unreferenced_files))


from . import PackageNotice, PackageWarning

class MissingPackageSupplementIssue(PackageNotice):
	title = 'Missing Package Supplement File'
	reference = 'http://docs.daz3d.com/doku.php/public/software/install_manager/referenceguide/tech_articles/package_supplement_file/'
	description = "A Package Supplement File 'Supplement.dsx' is recommended"

class InvalidPackageSupplementIssue(PackageWarning):
	title = 'Invalid Package Supplement File'
	reference = 'http://docs.daz3d.com/doku.php/public/software/install_manager/referenceguide/tech_articles/package_supplement_file/'
	description = "Package Supplement File 'Supplement.dsx' is invalid"

class EmbeddedPackageSupplementElementIssue(PackageWarning):
	title = 'Inexpedient Element in Embedded Package Supplement File'
	reference = 'http://docs.daz3d.com/doku.php/public/software/install_manager/referenceguide/tech_articles/package_supplement_file/'
	description = "Embedded Package Supplement File should not contain element"

class MismatchedPackageSupplementProductNameIssue(PackageNotice):
	title = 'Supplement File Product Name / Package Filename Mismatch'
	reference = 'http://docs.daz3d.com/doku.php/public/software/install_manager/referenceguide/tech_articles/package_naming/'
	description = "The package filename does not appear to be a simplified version of the Supplement File product name"

class IncorrectPackageSupplementProductStoreIDXIssue(PackageWarning):
	title = 'Incorrect ProductStoreIDX in Package Supplement File'
	reference = 'http://docs.daz3d.com/doku.php/public/software/install_manager/referenceguide/tech_articles/package_supplement_file/'
	description = "Package Supplement File contains ProductStoreIDX with incorrectly encoded source prefix"



from . import ProductNotice, ProductWarning

class InvalidProductMetadataIssue(ProductWarning):
	title = 'Invalid Metadata file'
	description = "Metadata file is invalid"

class MetadataFilenameIssue(ProductWarning):
	title = "Metadata filename does not match expected value"
	description = "Metadata filename does not match auto generated filename"

class PackageProductNotMetadataProductIssue(ProductNotice):
	title = "Package product name not in metadata product names"
	description = "There is no metadata product with the same name as the package product"

class MetadataFileContainsMultipleStoreIDsIssue(ProductWarning):
	title = "Metadata file contains multiple StoreIDs"
	description = "Metadata file contains products with differing StoreIDs"

class MetadataStoreIDIsDAZ3DIssue(ProductWarning):
	title = "Metadata StoreID is 'DAZ 3D'"
	description = "Only officially supplied DAZ 3D products should have a StoreID of 'DAZ 3D'"

class MetadataStoreIDIsNotDAZ3DIssue(ProductWarning):
	title = "Metadata StoreID is not 'DAZ 3D'"
	description = "Official DAZ 3D products should have a StoreID of 'DAZ 3D'"

class MetadataStoreIDIsNotLocalUserIssue(ProductNotice):
	title = "Metadata StoreID is not 'LOCAL USER'"
	description = "If StoreID does not already exist on users system, 'LOCAL USER' will be used instead"

class MetadataMissingProductTokenIssue(ProductWarning):
	title = "Metadata missing product token"
	description = "Metadata contains no product token for product with StoreID not set to 'LOCAL USER'"

class MetadataMissingArtistIssue(ProductWarning):
	title = "Metadata missing artist"
	description = "Metadata contains no artist entries"

class MetadataUnexpectedAssetsIssue(ProductWarning):
	title = "Unexpected Asset(s) in metadata"
	description = "Metadata contains Asset for non existent file"

class MetadataUnexpectedSupportAssetsIssue(ProductWarning):
	title = "Unexpected Support Asset(s) in metadata"
	description = "Metadata contains Support Asset for non existent file"

class MetadataUnlistedAssetsIssue(ProductWarning):
	title = "File(s) not listed as Assets or SupportAssets"
	description = "Metadata does not list all files as Assets or SupportAssets"

class MetadataInvalidSupportAssetFileRefIssue(ProductWarning):
	title = "Invalid metadata SupportAssets file reference"
	description = "Metadata SupportAssets value should refer to file reference is in"

class MetadataMissingContentTypeIssue(ProductWarning):
	title = "Missing Content Type for Assets(s) in metadata"
	reference = [
		'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/userguide/creating_content/packaging/tutorials/packaging_genesis_morph_products/start',
		'http://docs.daz3d.com/doku.php/public/dson_spec/format_description/metadata/content_types/start']
	description = "Metadata contains no Content Type for Asset"

class MetadataMissingAudienceIssue(ProductWarning):
	title = "Missing Audience for Assets(s) in metadata"
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/userguide/creating_content/packaging/tutorials/adding_metadata/start'
	description = "Metadata contains no Audience for Asset"

class MetadataNonStandardAudienceIssue(ProductWarning):
	title = "Non standard Audience for Assets(s) in metadata"
	description = "Metadata contains non standard Audience for Asset. Audience will be reassigned at import."

class MetadataMissingCategoriesIssue(ProductWarning):
	title = "Missing Categories for Assets(s) in metadata"
	reference = [
		'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/userguide/creating_content/packaging/tutorials/adding_metadata/start',
		'http://docs.daz3d.com/doku.php/public/dson_spec/format_description/metadata/categories/start']
	description = "Metadata contains no Categories for Asset"

# Note - Degrading Warning to Notice
# DAZ3D QA has recommendations for Tags to add, but does not follow this recommendation themselves
class MetadataMissingTagsIssue(ProductNotice):
	title = "Missing Tags for Assets(s) in metadata"
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/userguide/creating_content/packaging/tutorials/adding_metadata/start'
	description = "Metadata contains no Tags for Asset. DAZ QA recommends adding a number of search terms here, but rarely does so themselves."

class MetadataMissingCompatibilitiesIssue(ProductWarning):
	title = "Missing Compatibilities for Assets(s) in metadata"
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/userguide/creating_content/packaging/tutorials/adding_metadata/start'
	description = "Metadata contains no Compatibilities for Asset"

class MetadataMissingCompatibilityBaseIssue(ProductWarning):
	title = "Missing Compatibility Base for Assets(s) in metadata"
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/userguide/creating_content/packaging/tutorials/adding_metadata/start'
	description = "Metadata contains no Compatibility Base for Asset"

class MetadataHasDescriptionIssue(ProductNotice):
	title = "Description for Assets(s) in metadata"
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/userguide/creating_content/packaging/tutorials/adding_metadata/start'
	description = "Asset Description should usually be left blank"

class MetadataHasUserwordsIssue(ProductWarning):
	title = "Userwords for Assets(s) in metadata"
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/userguide/creating_content/packaging/tutorials/adding_metadata/start'
	description = "Asset Userwords are reserved for end users and should not be included in metadata"

class MetadataDuplicateCategoriesUsedIssue(ProductWarning):
	title = "Duplicate Categories for Assets(s) in metadata"
	description = "Metadata Categories for Asset should only be listed once per asset"

class MetadataParentCategoriesUsedIssue(ProductWarning):
	title = "Non most specific Categories for Assets(s) in metadata"
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/userguide/creating_content/packaging/tutorials/adding_metadata/start'
	description = "Metadata Categories for Asset should only list the most specific sub-categories"

class MetadataSpecialCategoriesUsedIssue(ProductWarning):
	title = "Special Categories for Assets(s) in metadata"
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/referenceguide/interface/panes/smart_content/products_page/category_view/start'
	description = "Metadata Categories for Asset should not include Special categories"

class MetadataUserCategoriesUsedIssue(ProductWarning):
	title = "User Categories for Assets(s) in metadata"
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/referenceguide/interface/panes/smart_content/products_page/category_view/start'
	description = "Metadata Categories for Asset should not include User categories"

class MetadataDeprecatedContentTypeIssue(ProductWarning):
	title = "Deprecated Content Type for Assets(s) in metadata"
	reference = 'http://docs.daz3d.com/doku.php/public/dson_spec/format_description/metadata/content_types/start'
	description = "Metadata contains deprecated Content Type for Asset"

class MetadataDeprecatedCategoriesIssue(ProductWarning):
	title = "Deprecated Categor(y|ies) for Assets(s) in metadata"
	reference = 'http://docs.daz3d.com/doku.php/public/dson_spec/format_description/metadata/categories/start'
	description = "Metadata contains deprecated Category for Asset"

class MetadataWrongShaderTypeIssue(ProductWarning):
	title = "Wrong Material or Shader type for Assets(s) in metadata"
	description = "Metadata contains the wrong Material or Shader type for Asset"

class MetadataNonspecificShaderTypeIssue(ProductNotice):
	title = "Nonspecific Material or Shader type for Assets(s) in metadata"
	description = "Metadata Content Type for Material and Shader Asset should specify the Shader type when possible"


from . import ProductWarning

class ExcessRuntimeLibrariesDirectoryIssue(ProductWarning):
	title = 'Runtime/Libraries contains sub "Libraries" directory'
	description = "Runtime/Libraries directory tree should be in root of Runtime/Libraries directory"

class FilesInRootOfRuntimeLibrariesDirectoryIssue(ProductWarning):
	title = 'Unexpected file(s) in root of Libraries directory'
	description = "No files should be in root of Libraries directory"

class UnexpectedDirectoriesInRuntimeLibrariesDirectoryIssue(ProductWarning):
	title = 'Unexpected directories(s) in Runtime/Libraries directory'
	description = "Non standard directory in Runtime/Libraries directory"

class InvalidCompressedFilesInRuntimeLibrariesDirectoryIssue(ProductWarning):
	title = 'Invalid compressed file(s) in Runtime/Libraries directory'
	description = "Invalid compressed file in Runtime/Libraries directory"


from . import PackageWarning

# NOTE: Using PackageWarning for plain ZIP archives....
class PathCollisionsInZipFile(PackageWarning):
	title = 'Path Collision(s) in ZIP file'
	description = "Path collisions found in ZIP file"


from abc import ABC
from collections.abc import Iterable

Detail = str | Exception | Iterable['Detail'] | tuple[str, 'Detail']

class ValidationIssue(ABC):
	"""Abstract base class for recording issues discovered during validation.

	Subclasses must set title and description.
	Subclasses may set reference.

	Arguments:
		detail (str | Iterable[str]): details of issue discovered
	"""

	type: str
	title: str
	description: str
	reference: str | list[str] = ''

	def __init__(self, detail: Detail | None = None) -> None:
		self.detail = detail

	def __str__(self) -> str:
		return f"{self.title}{f': {self.detail}' if self.detail else ''}\n{self.description}\n{self.reference}"

class ValidationNotice(ValidationIssue):
	"""Notice level issue."""

	type = 'NOTICE'

	def __str__ (self) -> str:
		return f'{type}: ' + super().__str__()

class ValidationWarning(ValidationIssue):
	"""Warning level issue."""

	type = 'WARNING'

	def __str__ (self) -> str:
		return f'{type}: ' + super().__str__()

class PackageNotice(ValidationNotice):
	"""Notice level package issue."""
	pass

class PackageWarning(ValidationWarning):
	"""Warning level package issue."""
	pass

class ProductNotice(ValidationNotice):
	"""Notice level product issue."""
	pass

class ProductWarning(ValidationWarning):
	"""Warning level product issue."""
	pass


from fs.errors import ResourceNotFound

from ..issues import supplementfile as issues
from ..validator.package import Package
from ..validator.ruleset import PackageRuleset, rule
from ..validator.schema import SchemaCheckedXML
from ..validator.utilities import alpha_numeric

class ValidateSupplementFiles(PackageRuleset):
	"""Perform supplement file validation of package.

	Arguments:
		data (ValidationData): validation data of package to validate.
	"""

	def _validate(self) -> None:
		"""Perform validation."""

		super()._validate()

		missing_supplement_files: list[str] = []
		invalid_supplement_files: dict[str, Exception] = {}
		self.embeded_supplement_files: dict[str, list[str]] = {}
		self.name_mismatch: dict[str, tuple[str, str]] = {}
		self.store_idx_mismatch: dict[str, tuple[str, str]] = {}
		for package in self.data.packages:
			try:
				self.supplement = SchemaCheckedXML(package.root_fs.openbin('Supplement.dsx'), 'Supplement.xsd').tree
			except ResourceNotFound:
				missing_supplement_files.append(package.path.as_posix())
			except Exception as e:
				invalid_supplement_files[package.path.as_posix()] = e
			else:
				self._checkPackageSupplementEmbeddedExclusionElements(package)
				self._checkPackageSupplementProductName(package)
				self._checkPackageSupplementProductStoreIDX(package)

		if missing_supplement_files:
			self._addIssue(issues.MissingPackageSupplementIssue(missing_supplement_files))

		if invalid_supplement_files:
			self._addIssue(issues.InvalidPackageSupplementIssue(invalid_supplement_files))

		if self.embeded_supplement_files:
			self._addIssue(issues.EmbeddedPackageSupplementElementIssue(self.embeded_supplement_files))

		if self.name_mismatch:
			self._addIssue(issues.MismatchedPackageSupplementProductNameIssue(self.name_mismatch))

		if self.store_idx_mismatch:
			self._addIssue(issues.IncorrectPackageSupplementProductStoreIDXIssue(self.store_idx_mismatch))

	@rule
	def _checkPackageSupplementEmbeddedExclusionElements(self, package: Package) -> None:
		"""Check for excluded elements in embedded package supplement file."""

		embeded_elements: list[str] = []
		for element_name in ['UserOrderId', 'UserOrderDate', 'InstallerDate', 'ProductFileGuid']:
			if len(self.supplement.xpath(f"/ProductSupplement/{element_name}")):
				embeded_elements.append(element_name)

		if embeded_elements:
			self.embeded_supplement_files[package.path.as_posix()] = embeded_elements

	@rule
	def _checkPackageSupplementProductName(self, package: Package) -> None:
		"""Check product name in package filename is simplified version of ProductName in package supplement file.

		Updates `package.product_name` to value of ProductName in package supplement file.
		"""

		package.product_name = self.supplement.xpath("/ProductSupplement/ProductName")[0].attrib['VALUE']

		if package.parsed and package.parsed['name'] != (simplified_name := alpha_numeric(package.product_name)):
			self.name_mismatch[package.path.as_posix()] = (simplified_name, package.parsed['name'])

	@rule
	def _checkPackageSupplementProductStoreIDX(self, package: Package) -> None:
		"""Check product and package ID in package filename matches ProductStoreIDX in package supplement file."""

		if package.parsed and (element := self.supplement.xpath("/ProductSupplement/ProductStoreIDX")):
			if len(element) and element[0].attrib['VALUE'] != package.product_store_idx:
				self.store_idx_mismatch[package.path.as_posix()] = (element[0].attrib['VALUE'], package.product_store_idx or '')


import re

from enum import Enum
from hashlib import md5
from pathlib import Path
from typing import cast, Self, TypedDict
from uuid import UUID

from fs.dimzipfs import DIMZipFS

from .resources import read_list_from

class PackageType(Enum):
	"""Whether package is a standard or custom package."""

	STANDARD = 0
	CUSTOM = 1

class ParsedPackageFilename(TypedDict):
	"""Data extracted from package filename."""

	prefix: str
	sku: str
	id: str | None
	name: str

class Package:
	"""Data for a package.

	Arguments:
		package (DIMZipFS): `DIMZipFS` pyfilesystem2 filesystem
	"""

	def __init__(self, package: DIMZipFS) -> None:
		path = package.delegate_fs()._file  # pyright: ignore[reportPrivateUsage]
		if not isinstance(path, str):
			raise ValueError

		self.path = Path(path)
		self.root_fs = package.delegate_fs()
		self.parsed: ParsedPackageFilename | None = None
		self.type: PackageType | None = None
		self.product_name: str | None = None
		self.product_store_idx: str | None = None
		self.product_file_guid = self._calculateProductFileGUID()

		if self.path and (matches := re.match(r'^(?P<prefix>[A-Z][0-9A-Z]{0,6})(?=\d{8})(?P<sku>\d{8})(-(?P<id>\d{2}))?_(?P<name>[0-9A-Za-z]+)\.zip$', self.path.name)):
			self.parsed = cast(ParsedPackageFilename, matches.groupdict())

		if self.parsed:
			self.type = PackageType(self.parsed['prefix'] not in read_list_from('daz/reserved_package_prefixes.txt'))

			if self.type == PackageType.STANDARD:
				self.product_store_idx = f"{self.parsed['sku'].lstrip('0')}-{self.parsed['id']}"
			else:
				self.product_store_idx = f"{str(int(self.parsed['prefix'], 36))}{self.parsed['sku']}-{self.parsed['id']}"

	def __lt__(self, other: Self) -> bool:
		return self.path.name < other.path.name

	def _calculateProductFileGUID(self) -> UUID:
		"""Calculate `ProductFileGUID` used in DIM generated `Supplement.dsx` stored outside of product zip."""

		hash = md5()

		with open(self.path, 'rb') as file:
			while chunk := file.read(hash.digest_size * hash.block_size * 8):
				hash.update(chunk)

		return UUID(bytes = hash.digest())


from . import ProductWarning

class MissingMetadataFilesIssue(ProductWarning):
	title = 'Missing Metadata file(s)'
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/userguide/creating_content/packaging/tutorials/adding_metadata/start'
	description = "Product should provide a metadata file"

class SubdirectoriesInRuntimeSupportIssue(ProductWarning):
	title = 'Unexpected subdirector(y|ies) in Runtime/Support'
	description = "Runtime/Support should not contain subdirectories"

class UnexpectedFilesInRuntimeSupportIssue(ProductWarning):
	title = 'Unexpected file(s) in Runtime/Support'
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/referenceguide/interface/panes/smart_content/products_page/results_view/'
	description = "Files in Runtime/Support should have the same stem as the DSX file providing metadata"

class MissingMetadataIconFilesIssue(ProductWarning):
	title = 'Missing Metadata Icon file(s)'
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/referenceguide/interface/panes/smart_content/products_page/results_view/'
	description = "Product should provide an icon with metadata"

class MissingMetadataScriptFilesIssue(ProductWarning):
	title = 'Missing Metadata Script file(s)'
	description = "Product should provide a default script with metadata"

class RedundantMetadataIconFilesIssue(ProductWarning):
	title = 'Redundant Metadata Icon file(s)'
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/change_log_4_9_4_122#4_9_4_31'
	description = "Metadata Icon files should not be provided in multiple formats"

class UndersizedMetadataIconFilesIssue(ProductWarning):
	title = 'Undersized Metadata Icon file(s)'
	reference = [
		'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/userguide/creating_content/packaging/tutorials/adding_metadata/start',
		'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/referenceguide/interface/panes/smart_content/products_page/results_view/start']
	description = "Metadata Icon files should be 114x148 pixels or larger"


import re

from collections import defaultdict
from fs.path import splitext
from itertools import chain
from pathlib import Path
from typing import cast

from ..issues import contentdirectory as issues
from ..validator.resources import read_list_from
from ..validator.ruleset import ProductRuleset, rule
from ..validator.schema import SchemaCheckedJSON
from ..validator.utilities import Step, checkTypo, thumbnailsFor, trackDependencyIfExists

_TEXTURES_DIR = 'Runtime/Textures'
_NON_USER_FACING_DIRECTORIES = [d.lower() for d in read_list_from('daz/user_facing_excluded_directories.txt')]
_USER_FACING_FILE_EXTENSIONS = [f'*.{ext.lower()}' for ext in read_list_from('daz/native_file_extensions.txt')]


class ValidateContentDirectory(ProductRuleset):
	"""Perform validation of product.

	Arguments:
		data (ValidationData): validation data of product to validate.
	"""

	def _validate(self) -> None:
		"""Perform validation."""

		super()._validate()
		self._checkCaseInsensitivePathCollisions()
		self._checkExcessContentDirectory()
		self._checkEmptyDirectories()
		self._checkRootDirectories()
		self._checkRootFiles()
		self._checkGratuitousFiles()
		self._checkEmptyFiles()
		self._checkDazJsonLinkFiles()
		self._checkLegacyFormatFiles()
		self._checkAssetThumbnails()
		self._checkUnexpectedUserFacingFiles()

		# data.referenced_files must be populated by ValidateDSONFiles, etc. before this step
		self._checkUnreferencedTextureFiles()

		# data.missing_referenced_files must be populated by ValidateDSONFiles, etc. before this step
		self._checkInvalidFileReferences()

		# data.vendor_paths must be populated by ValidateRuntimeDirectory, ValidateDataDirectory, etc. before this step
		self._checkDazVendorDirectories()

	@rule
	def _checkCaseInsensitivePathCollisions(self) -> None:
		"""Check product for case insensitive path collisions."""

		case_insensitive_path_count: dict[str, list[str]] = defaultdict(list)
		for entry in self.data.product_fs_unwrapped.walk.info():  # pyright: ignore[reportUnknownMemberType]
			case_insensitive_path_count[entry[0].lower()].append(entry[0])
		
		if case_insensitive_collisions := [entries for entries in case_insensitive_path_count.values() if len(entries) > 1]:
			self._addIssue(issues.CaseInsensitivePathCollision(case_insensitive_collisions))

	@rule
	def _checkExcessContentDirectory(self) -> None:
		"""Check product for content directory in content directory."""

		if self.data.product_fs.isdir('Content'):
			self._addIssue(issues.ExcessContentDirectoryIssue(['Content/Content']))

	@rule
	def _checkEmptyDirectories(self) -> None:
		"""Check product for empty directories."""

		empty_dirs: set[str] = set()
		for directory in self.data.product_fs.walk(search='depth'):  # pyright: ignore[reportUnknownMemberType]
			directory = cast(Step, directory)
			if not directory.files and (not directory.dirs or
				set(((Path(directory.path) / d.name).as_posix() for d in directory.dirs)).issubset(empty_dirs)):
				empty_dirs.add(Path(*Path(directory.path).parts[1:]).as_posix())

		if empty_dirs:
			self._addIssue(issues.EmptyDirectoriesIssue(empty_dirs))

	@rule
	def _checkRootDirectories(self) -> None:
		"""Check product for uncommon directories in root content directory."""

		expected_directories = [d for d in chain(read_list_from('daz/common_content_directories.txt'), read_list_from('daz/user_facing_excluded_directories.txt'))]
		expected_directories_lc = [d.lower() for d in expected_directories]

		unexpected_directory_alternatives: dict[str, list[str]] = {}
		if unexpected_directories := [entry.name for entry in self.data.product_fs.scandir('') if entry.is_dir and not entry.name.lower() in expected_directories_lc]:
			for dir in unexpected_directories:
				matches = checkTypo(dir, expected_directories, 0.5)
				unexpected_directory_alternatives[dir] = matches if matches and dir.lower() != 'content' else []

		if unexpected_directory_alternatives:
			self._addIssue(issues.UncommonDirectoryInRootOfContentDirectoryIssue(unexpected_directory_alternatives))

	@rule
	def _checkRootFiles(self) -> None:
		"""Check product for files in root content directory."""

		if root_files := [entry.name for entry in self.data.product_fs.scandir('') if entry.is_file]:
			self._addIssue(issues.FilesInRootOfContentDirectoryIssue(root_files))

	@rule
	def _checkGratuitousFiles(self) -> None:
		"""Check product for gratuitous files."""

		GRATUITOUS_FILES = ['Thumbs.db', 'InstallManagerFileRegister.json']

		gratuitous_files: list[str] = []
		for file in (file.lstrip('/') for file in self.data.product_fs.walk.files()):  # pyright: ignore[reportUnknownMemberType]
			path = Path(file)
			if path.name.startswith('.') or '__MACOSX' in path.parts or path.name.lower() in [x.lower() for x in GRATUITOUS_FILES]:
				gratuitous_files.append(file)

		for directory in (dir.lstrip('/') for dir in self.data.product_fs.walk.dirs()):  # pyright: ignore[reportUnknownMemberType]
			path = Path(directory)
			if path.name.startswith('.') or '__MACOSX' in path.parts:
					gratuitous_files.append(directory)

		if gratuitous_files:
			self._addIssue(issues.GratuitousFilesIssue(gratuitous_files))

	@rule
	def _checkEmptyFiles(self) -> None:
		"""Check product for empty files."""

		# NOTE: There are valid use cases for 0 byte files such as Instructions.duf with Instructions.png being the relevant part
		# May need to modify this check

		if empty_files := [file.lstrip('/') for file, info in self.data.product_fs.walk.info(namespaces=['details']) if info.is_file and info.size == 0]:  # pyright: ignore[reportUnknownMemberType]
			self._addIssue(issues.EmptyFilesIssue(empty_files))

	@rule
	def _checkDazJsonLinkFiles(self) -> None:
		"""Check DAZ JSON Link files."""

		invalid_djl_files: dict[str, Exception] = {}
		unnecessary_thumbnails_for_djl_files: dict[str, list[str]] = {}
		for file in (file.lstrip('/') for file in self.data.product_fs.walk.files(filter=['*.djl'])):  # pyright: ignore[reportUnknownMemberType]
			thumbnails = [t for lst in thumbnailsFor(self.data.product_fs, file) for t in lst]

			try:
				json = SchemaCheckedJSON(self.data.product_fs.openbin(file), 'djl.schema.json')
			except Exception as e:
				invalid_djl_files[file] = e
			else:
				# NOTE: DS writes .djl file paths as all lowercase, so check filesystem as case insensitive
				if not trackDependencyIfExists(self.data, json.data['path'], file):
					self.data.missing_referenced_files.setdefault(json.data['path'], set()).add(file)

				if thumbnails:
					target_thumbnails = [t for lst in thumbnailsFor(self.data.filesystem, json.data['path']) for t in lst]
					if target_thumbnails:
						unnecessary_thumbnails_for_djl_files[file] = thumbnails

		if invalid_djl_files:
			self._addIssue(issues.InvalidDJLFilesIssue(invalid_djl_files))

		if unnecessary_thumbnails_for_djl_files:
			self._addIssue(issues.UnnecessaryThumbnailsForDJLIssue(unnecessary_thumbnails_for_djl_files))


	@rule
	def _checkLegacyFormatFiles(self) -> None:
		"""Check for legacy serialized binary files."""

		# http://docs.daz3d.com/doku.php/public/software/dazstudio/4/referenceguide/tech_articles/file_extensions/start#serialized_binary
		LEGACY_FILE_EXTENSIONS = ['*.daz', '*.ds']

		if legacy_format_files := [entry.lstrip('/') for entry in self.data.product_fs.walk.files(filter=LEGACY_FILE_EXTENSIONS)]:  # pyright: ignore[reportUnknownMemberType]
			self._addIssue(issues.LegacyFilesIssue(legacy_format_files))

	@rule
	def _checkDazVendorDirectories(self) -> None:
		"""Check DAZ vendor directories."""

		if not self.data.daz_original and (daz_vendor_dirs := [dir for vendor in self.data.vendor_paths.keys() for dir in self.data.vendor_paths[vendor] if vendor.lower().startswith('daz')]):
			self._addIssue(issues.FilesInVendorDazDirectoryIssue(daz_vendor_dirs))
		
	@rule
	def _checkInvalidFileReferences(self) -> None:
		"""Check invalid file references."""

		if self.data.missing_referenced_files:
			absoluteRegex = re.compile(r'^\/[a-zA-Z]:\/')

			absoluteFileReferences: dict[str, set[str]] = {}
			missingFileReferences: dict[str, set[str]] = {}
			for k, v in self.data.missing_referenced_files.items():
				if re.match(absoluteRegex, k):
					absoluteFileReferences[k] = v
				else:
					missingFileReferences[k] = v

			if absoluteFileReferences:
				self._addIssue(issues.FilesReferenceAbsolutePathsIssue(absoluteFileReferences))

			if missingFileReferences:
				self._addIssue(issues.FilesReferenceNonexistentFilesIssue(missingFileReferences))

	@rule
	def _checkUnreferencedTextureFiles(self) -> None:
		"""Check product for unreferenced files in Runtime/Textures directory."""
		# NOTE: This test is here instead of in ValidateRuntimeDirectory due to test dependencies.

		if self.data.product_fs.isdir(_TEXTURES_DIR):
			referenced_files_lc = {f.lower() for f in self.data.referenced_files}
			if unreferenced_files := [file for file in self.data.product_fs.walk.files(_TEXTURES_DIR) if not '/' + file.lower() in referenced_files_lc]:  # pyright: ignore[reportUnknownMemberType]
				self._addIssue(issues.UnreferencedFilesInTexturesDirectoryIssue(unreferenced_files))

	@rule
	def _checkAssetThumbnails(self) -> None:
		"""Check product thumbnails."""

		missingThumbnails: list[str] = []
		fullExtensionThumbnails: list[str] = []

		for file in (file.lstrip('/') for file in self.data.product_fs.walk.files(filter=_USER_FACING_FILE_EXTENSIONS, exclude_dirs=_NON_USER_FACING_DIRECTORIES)):  # pyright: ignore[reportUnknownMemberType]
			fullExtThumbnail = f'{file}.png'

			if not self.data.product_fs.exists(fullExtThumbnail):
				if not self.data.product_fs.exists(f'{splitext(file)[0]}.png'):
					missingThumbnails.append(file)
			else:
				fullExtensionThumbnails.append(fullExtThumbnail)

		if missingThumbnails:
			self._addIssue(issues.MissingThumbnailsIssue(missingThumbnails))

		if fullExtensionThumbnails:
			self._addIssue(issues.FullExtensionThumbnailsIssue(fullExtensionThumbnails))

	@rule
	def _checkUnexpectedUserFacingFiles(self) -> None:
		"""Check for unexpected files in user facing directories."""

		unexpectedFiles: list[str] = []
		fullExtensionTipFiles: list[str] = []

		for file in (file.lstrip('/') for file in self.data.product_fs.walk.files(exclude=_USER_FACING_FILE_EXTENSIONS + ['*.djl'], exclude_dirs=_NON_USER_FACING_DIRECTORIES)):  # pyright: ignore[reportUnknownMemberType]
			if file.lower().endswith('.png'):
				splitFile = splitext(file)[0]
				if not self.data.product_fs.exists(splitFile):
					if not any(self.data.product_fs.exists(f) for f in (splitFile + e[1:] for e in _USER_FACING_FILE_EXTENSIONS + ['*.djl'])):
						if splitFile.lower().endswith('.tip'):
							splitFile = splitext(splitFile)[0]
							if self.data.product_fs.exists(splitFile):
								fullExtensionTipFiles.append(file)
							else:
								if not any(self.data.product_fs.exists(f) for f in (splitFile + e[1:] for e in _USER_FACING_FILE_EXTENSIONS + ['*.djl'])):
									unexpectedFiles.append(file)
						else:
							unexpectedFiles.append(file)
			else:
				unexpectedFiles.append(file)

		if unexpectedFiles:
			self._addIssue(issues.UnexpectedFilesInUserFacingDirectoriesIssue(unexpectedFiles))

		if fullExtensionTipFiles:
			self._addIssue(issues.FullExtensionTipFilesIssue(fullExtensionTipFiles))


from . import ProductWarning, ProductNotice

class InvalidDSONFilesIssue(ProductWarning):
	title = 'Invalid DSON file(s) in Content directory'
	reference = 'http://docs.daz3d.com/doku.php/public/dson_spec/start'
	description = "Content directory contains invalid DSON file"

class AssetIDMismatchFilesIssue(ProductNotice):
	title = 'Asset ID mismatch(es) in DSON file(s)'
	description = "Asset ID in DSON file does not match filename"

class GeometryInDUFFilesIssue(ProductWarning):
	title = 'Geometry in DSON User File(s) (*.duf)'
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/userguide/creating_content/packaging/tutorials/packaging_genesis_morph_products/start'
	description = "Geometry should be saved as Support Assets, to prevent being resaved in user's Scene file"

class UVSetInDUFFilesIssue(ProductWarning):
	title = 'UV Set(s) in DSON User File(s) (*.duf)'
	reference = 'http://docs.daz3d.com/doku.php/public/publishing/uv_update_replace/start'
	description = ("UV sets should be saved as Support Assets, to prevent being resaved in user's Scene file. If this item is also listed "
	               "as having a Geometry in DSON User File(s) issue, solving that issue should also solve this issue.")

class MorphInDUFFilesIssue(ProductWarning):
	title = 'Morph(s) in DSON User File(s) (*.duf)'
	reference = 'http://docs.daz3d.com/doku.php/public/software/dazstudio/4/userguide/creating_content/packaging/tutorials/saving_morphs/start#custom_or_dialed_morphs'
	description = "Morphs should be saved as Support Assets, to prevent being resaved in user's Scene file"

class ShaderInDUFFilesIssue(ProductWarning):
	title = 'Custom Shader(s) in DSON User File(s) (*.duf)'
	description = "Custom Shaders should be saved as Support Assets, to prevent being resaved in user's Scene file"

class ActiveMorphsInDSFFilesIssue(ProductWarning):
	title = 'Active Morph(s) in DSON Support File(s) (*.dsf)'
	description = "Morphs should be set to 0 or False when saved as Support Assets. Use Presets to set active values."


from collections import defaultdict
from os.path import normpath
from zipfile import ZipFile

from ..issues import zipfile as issues
from ..validator.ruleset import Ruleset, rule

class ValidateZipFiles(Ruleset):
	"""Perform zip file validation of product.

	Arguments:
		data (ValidationData): validation data of product to validate.
	"""

	def _validate(self) -> None:
		"""Perform validation."""

		super()._validate()

		self.collisions: dict[str, list[list[str]]] = {}
		for zip in self.data.zips:
			self._checkPathCollisions(zip._zip)  # pyright: ignore[reportPrivateUsage]

		if self.collisions:
			self.data.issues.package.append(issues.PathCollisionsInZipFile(self.collisions))

	@rule
	def _checkPathCollisions(self, zip: ZipFile) -> None:
		"""Check product zip file for path collisions.

		A ZIP archive may contain `a/b`, `a/b`, and `a\\b`, and still be valid
		but this will not extract in a useful way with most tools.
		"""

		path_count: dict[str, list[str]] = defaultdict(list)
		for entry in zip.namelist():
			path_count[normpath(entry)].append(entry)
		
		if path_collisions := [entries for entries in path_count.values() if len(entries) > 1]:
			self.collisions[zip.filename or 'Unknown'] = path_collisions


from . import ProductWarning

class ExcessGeometriesDirectoryIssue(ProductWarning):
	title = 'Runtime/Geometries contains sub "geometries" directory'
	description = "Runtime/Geometries directory tree should be in root of Runtime/Geometries directory"

class FilesInRootOfGeometriesDirectoryIssue(ProductWarning):
	title = 'Unexpected file(s) in root of Geometries directory'
	description = "No files should be in root of Runtime/Geometries directory"

class UnexpectedFilesInRuntimeGeometriesDirectoryIssue(ProductWarning):
	title = 'Unexpected file(s) in Runtime/Geometries directory'
	description = "Only wavefront *.obj and *.mtl files should be in Runtime/Geometries directory"

class InvalidCompressedFilesInRuntimeGeometriesDirectoryIssue(ProductWarning):
	title = 'Invalid compressed file(s) in Runtime/Geometries directory'
	description = "Invalid compressed file in Runtime/Geometries directory"


from . import ProductNotice, ProductWarning

class ExcessDataDirectoryIssue(ProductWarning):
	title = 'Data contains sub "data" directory'
	description = "Data directory tree should be in root of data directory"

class FilesInRootOfDataDirectoryIssue(ProductWarning):
	title = 'Unexpected file(s) in root of data directory'
	description = "Files in data directory should be in 'data/Vendor/Product/Item/'"

class FilesInDataVendorDirectoryIssue(ProductWarning):
	title = 'Unexpected file(s) in data/Vendor directory'
	description = "Files in data directory should be in 'data/Vendor/Product/Item/'"

class FilesInDataProductDirectoryIssue(ProductWarning):
	title = 'Unexpected file(s) in data/Vendor/Product directory'
	description = "Files in data directory should be in 'data/Vendor/Product/Item/'"

class AutoAdaptedInDataDirectoryIssue(ProductWarning):
	title = 'auto_adapted directory in data directory'
	description = "Assets loaded from non-native formats should be resaved as Support Assets"
	reference = 'https://www.daz3d.com/forums/discussion/comment/604443/#Comment_604443'

class LegacyDirectoriesInDataDirectoryIssue(ProductNotice):
	title = 'Legacy director(y|ies) (3_0, 4_0, 4_0_2) in data directory'
	description = "Support Assets in legacy formats should be resaved as DSON Support Assets with correct Vendor/Product/Item except when maintaining compatibility with previously distributed items"

class LegacyFilesInDataDirectoryIssue(ProductNotice):
	title = 'Legacy file(s) (*.dsd, *.dso, *.dsv) in data directory'
	description = "Support Assets in legacy formats should be resaved as DSON Support Assets except when maintaining compatibility with previously distributed items"

class DufFilesInDataDirectoryIssue(ProductNotice):
	title = 'DSON User File(s) (*.duf) in data directory without PostLoad reference'
	description = ("DSON User Files in data directory cannot be accessed by users directly but may be accessed via another method such as a script")

class UnreferencedFilesInDataDirectoryIssue(ProductNotice):
	title = 'Unreferenced file(s) in data directory'
	description = ("Files in data directory cannot be accessed by users directly. Validation can detect references from DSON user files (*.duf) in most cases, "
	               "but not by scripts (*.dsa, *.dsb, *.dse), legacy files (*.daz, *.ds), or as modifiers to other products. Legacy file types (*.dsd, *.dso, *.dsv) and "
	               "files in the standard directory structure '/data/Vendor/Product/Item/{add-ons, morphs, uv sets, projection morphs, projection templates, tools}/Vendor/*/' "
	               "are ignored for this test. There may be false positives in this list.")


from ..issues import package as issues
from ..validator.package import Package, PackageType
from ..validator.ruleset import PackageRuleset, rule

class ValidatePackages(PackageRuleset):
	"""Perform package validation of product.

	Arguments:
		data (ValidationData): validation data of package to validate.
	"""

	def _validate(self) -> None:
		"""Perform validation."""

		super()._validate()

		self.bad_package_names: list[str] = []
		self.bad_custom_pkgs: list[str] = []
		self.bad_standard_pkgs: list[str] = []
		for package in self.data.packages:
			self._checkPackageNameParseable(package)
			self._checkDAZPackageType(package)
			self._checkOtherPackageType(package)

		if self.bad_package_names:
			self._addIssue(issues.PackageNameIssue(self.bad_package_names))

		if self.bad_custom_pkgs:
			self._addIssue(issues.CustomPackageNameIssue(self.bad_custom_pkgs))

		if self.bad_standard_pkgs:
			self._addIssue(issues.StandardPackageNameIssue(self.bad_standard_pkgs))

	@rule
	def _checkPackageNameParseable(self, package: Package) -> None:
		"""Check package filename is parseable."""

		if not package.parsed:
			self.bad_package_names.append(package.path.name)

	@rule
	def _checkDAZPackageType(self, package: Package) -> None:
		"""Check DAZ distributed package type is STANDARD."""

		if self.data.daz and package.type != PackageType.STANDARD:
			self.bad_custom_pkgs.append(package.path.name)

	@rule
	def _checkOtherPackageType(self, package: Package) -> None:
		"""Check non DAZ distributed package type is CUSTOM."""

		if not self.data.daz and package.type != PackageType.CUSTOM:
			self.bad_standard_pkgs.append(package.path.name)


from . import ProductWarning, ProductNotice

class CaseInsensitivePathCollision(ProductWarning):
	title = 'Case Insensitive Path Collision(s)'
	description = "Path collisions found for case insensitive file systems"

class FilesInRootOfContentDirectoryIssue(ProductWarning):
	title = 'Unexpected file(s) in root of Content directory'
	description = "No files should be in root of content directory"

class ExcessContentDirectoryIssue(ProductWarning):
	title = 'Content contains sub "content" directory'
	description = "Content directory tree should be in root of Content directory"

class EmptyDirectoriesIssue(ProductWarning):
	title = 'Empty director(y|ies) in Content directory'
	description = "Content directory tree includes directory with no files"

class GratuitousFilesIssue(ProductWarning):
	title = 'Gratuitous file(s) in Content directory'
	description = "Content directory includes files which are likely gratuitous"

class EmptyFilesIssue(ProductWarning):
	title = 'Empty file(s) in Content directory'
	description = "Content directory contains files with no content"

class InvalidDJLFilesIssue(ProductWarning):
	title = 'Invalid Daz JSON Link file(s) in Content directory'
	description = "Content directory contains invalid *.djl file"

class UnnecessaryThumbnailsForDJLIssue(ProductWarning):
	title = 'Unnecessary thumbnails exist for Daz JSON Link file(s) in Content directory'
	description = "Content directory contains thumbnail for *.djl file where *.djl target thumbnail exists"

class UncommonDirectoryInRootOfContentDirectoryIssue(ProductWarning):
	title = 'Uncommon director(y|ies) in root of Content directory'
	description = "Directory may be spelled or located incorrectly. Common near matches will be listed when possible."

class FilesInVendorDazDirectoryIssue(ProductWarning):
	title = 'Unexpected file(s) in reserved Vendor directory'
	description = "No files should be in directories for vendor DAZ"

class FilesReferenceNonexistentFilesIssue(ProductWarning):
	title = 'Invalid file reference(s)'
	description = "File references nonexistent files. This may indicate a missing dependency while validating."

class FilesReferenceAbsolutePathsIssue(ProductWarning):
	title = 'Absolute file path reference(s)'
	description = "File references should be relative to Content directory"

class LegacyFilesIssue(ProductWarning):
	title = 'Legacy file(s) (*.daz, *.ds) in Content directory'
	description = "Assets in legacy formats should be resaved as DSON Assets"

class MissingThumbnailsIssue(ProductWarning):
	title = 'Missing thumbnail(s)'
	description = "Assets should have thumbnails"

class FullExtensionThumbnailsIssue(ProductWarning):
	title = 'Full extension thumbnail(s)'
	description = "Distributed assets should have asset extension removed from thumbnail name to prevent user from overwriting when resaving"

class UnexpectedFilesInUserFacingDirectoriesIssue(ProductWarning):
	title = 'Unexpected file(s) in user facing directory'
	description = ("Files in user facing directories should consist of *.duf, *.dsa, *.dsb, *.dse and accompanying thumbnails (*.png) and tips (*.tip.png). "
	               "These files may be misplaced or better located in the data, ReadMe's, or Runtime directories.")

class FullExtensionTipFilesIssue(ProductWarning):
	title = 'Full extension tip file(s)'
	description = "Tip files must not include the extension of the accompanying asset file"

class UnreferencedFilesInTexturesDirectoryIssue(ProductNotice):
	title = 'Unreferenced file(s) in Textures directory'
	description = ("Files in Runtime/Textures directory should be referenced by a user facing file. Validation can detect references from most DSON user files (*.duf) in most cases, "
	               "but not by scripts (*.dsa, *.dsb, *.dse) or legacy files (*.daz, *.ds). There may be false positives in this list.")


from .validationdata import ValidationData
from ..rules.datadirectory import ValidateDataDirectory
from ..rules.dsonfiles import ValidateDSONFiles
from ..rules.metadatafiles import ValidateMetadataFiles
from ..rules.package import ValidatePackages
from ..rules.packagemanifest import ValidatePackageManifests
from ..rules.packagesupplementfile import ValidateSupplementFiles
from ..rules.contentdirectory import ValidateContentDirectory
from ..rules.runtimedirectory import ValidateRuntimeDirectory
from ..rules.runtimelibrariesdirectory import ValidateRuntimeLibrariesDirectory
from ..rules.runtimegeometriesdirectory import ValidateRuntimeGeometriesDirectory
from ..rules.runtimesupportdirectory import ValidateRuntimeSupportDirectory
from ..rules.zipfile import ValidateZipFiles

def validate(data: ValidationData) -> None:
	rules = [
		ValidateZipFiles,
		ValidatePackages,
		ValidatePackageManifests,
		ValidateSupplementFiles,
		ValidateRuntimeSupportDirectory,
		ValidateRuntimeDirectory,
		ValidateRuntimeGeometriesDirectory,
		ValidateRuntimeLibrariesDirectory,
		ValidateDSONFiles,
		ValidateDataDirectory,	# data.dependency_files, data.postload_files must be populated by ValidateDSONFiles, etc. before this step
		ValidateContentDirectory,	# data.vendor_paths, data.referenced_files, data.missing_referenced_files must be populated by ValidateRuntimeDirectory, ValidateDataDirectory, ValidateDSONFiles, etc. before this step
		ValidateMetadataFiles,	# data.shader_users must be populated by ValidateDSONFiles, etc. before this step
	]

	for r in rules:
		r(data)


from . import ProductNotice, ProductWarning

class ExcessRuntimeDirectoryIssue(ProductWarning):
	title = 'Runtime contains sub "runtime" directory'
	description = "Runtime directory tree should be in root of Runtime directory"

class ExcessTexturesDirectoryIssue(ProductWarning):
	title = 'Runtime/Textures contains sub "textures" directory'
	description = "Runtime/Textures directory tree should be in root of Runtime/Textures directory"

class ExcessTemplatesDirectoryIssue(ProductWarning):
	title = 'Runtime/Templates contains sub "templates" directory'
	description = "Runtime/Templates directory tree should be in root of Runtime/Templates directory"

class FilesInRootOfRuntimeDirectoryIssue(ProductWarning):
	title = 'Unexpected file(s) in root of Runtime directory'
	description = "No files should be in root of Runtime directory"

class UnexpectedDirectoriesInRuntimeDirectoryIssue(ProductWarning):
	title = 'Unexpected directories(s) in Runtime directory'
	description = "Non standard directory in Runtime directory"

class FilesInRootOfTexturesDirectoryIssue(ProductWarning):
	title = 'Unexpected file(s) in root of Textures directory'
	description = "No files should be in root of Runtime/Textures directory"

class FilesInRootOfTemplatesDirectoryIssue(ProductWarning):
	title = 'Unexpected file(s) in root of Templates directory'
	description = "No files should be in root of Runtime/Templates directory"

class NonImageFilesInTexturesDirectoryIssue(ProductWarning):
	title = 'Unexpected non image file(s) in Textures directory'
	description = "Only image files should be in Runtime/Textures/Vendor directory"

class NonImageFilesInTemplatesDirectoryIssue(ProductWarning):
	title = 'Unexpected non image file(s) in Templates directory'
	description = "Only image files should be in Runtime/Templates/Vendor directory"

class AtypicalImageFilesInTexturesDirectoryIssue(ProductNotice):
	title = 'Atypical image file type(s) in Textures directory'
	description = "Consider using PNG, JPEG or OpenEXR"

class AtypicalImageFilesInTemplatesDirectoryIssue(ProductNotice):
	title = 'Atypical image file type(s) in Templates directory'
	description = "Consider using SVG, PNG, or JPEG"

class UnreadableImageFilesInTexturesDirectoryIssue(ProductWarning):
	title = 'Unreadable image file(s) in Textures directory'
	description = "Image file is invalid"

class UnreadableImageFilesInTemplatesDirectoryIssue(ProductWarning):
	title = 'Unreadable image file(s) in Templates directory'
	description = "Image file is invalid"

class ImageHasIncorrectFileExtensionIssue(ProductWarning):
	title = 'Image(s) with incorrect file extension'
	description = "Image file extension does not match detected image type"

class UnexpectedFilesInWebLinksDirectoryIssue(ProductWarning):
	title = 'Unexpected file(s) in WebLinks directory'
	description = "Only *.pzs files should be in WebLinks directory"

class InvalidPZSFileIssue(ProductWarning):
	title = 'Invalid WebLink file(s) in WebLinks directory'
	description = "WebLinks directory contains *.pzs file with invalid URL"

class WebLinksIssue(ProductWarning):
	title = 'Unexpected WebLinks directory'
	description = "WebLinks should only be provided for Poser content"


import json
import jsonschema

from pathlib import Path
from typing import cast, Any, BinaryIO

from fs.iotools import RawWrapper
from lxml import etree

from .resources import read_datafile_bytes

class SchemaCheckedXML():
	"""XML Schema validator

	Arguments:
		xml_file (pathlib.Path | typing.BinaryIO): XML file to validate
		schema (str): Schema name to validate against

	Raises:
		`lxml.etree.DocumentInvalid`
		`FileNotFoundError`
		...
	"""

	cache: dict[str, etree.XMLSchema] = {}

	def __init__(self, xml_file: Path | BinaryIO, schema: str) -> None:
		if not schema in self.cache:
			xmlschema = etree.XMLSchema(etree.fromstring(read_datafile_bytes(f'schemas/{schema}')))
			self.cache[schema] = xmlschema
		else:
			xmlschema = self.cache[schema]

		base_url = xml_file.name if isinstance(xml_file, (BinaryIO, RawWrapper)) else None
		self.tree = etree.parse(xml_file, base_url = base_url)

		xmlschema.assertValid(self.tree)

class SchemaCheckedJSON():
	"""JSON Schema validator

	Arguments:
		json_file (pathlib.Path | typing.BinaryIO): JSON file to validate
		schema (str): Schema name to validate against

	Raises:
		`json.JSONDecodeError`
		`jsonschema.exceptions.ValidationError`
		`jsonschema.exceptions.SchemaError`
		`FileNotFoundError`
		...
	"""

	cache: dict[str, jsonschema.Validator] = {}
	data: dict[str, Any]

	def __init__(self, json_file: Path | BinaryIO, schema: str) -> None:
		if not schema in self.cache:
			schema_data: dict[str, Any] = json.loads(read_datafile_bytes(f'schemas/{schema}'))
			validator = cast(jsonschema.Validator, jsonschema.validators.validator_for(schema_data))  # pyright: ignore[reportGeneralTypeIssues, reportUnknownMemberType]
			validator.check_schema(schema_data)
			self.cache[schema] = validator(schema_data)  # pyright: ignore[reportGeneralTypeIssues, reportUnknownArgumentType]

		validator = self.cache[schema]

		if isinstance(json_file, Path):
			with open(json_file, 'r') as file:
				self.data = json.load(file)
		else:
			self.data = json.load(json_file)

		validator.validate(self.data)


import json
import re

from jsonpath_ng.ext import parse

from typing import Any
from urllib.parse import urlparse, unquote
from warnings import warn

from ..issues import dsonfiles as issues
from ..validator.resources import read_list_from
from ..validator.ruleset import ProductRuleset, rule
from ..validator.schema import SchemaCheckedJSON
from ..validator.utilities import decompressDSON, trackDependencyIfExists

class ValidateDSONFiles(ProductRuleset):
	"""Perform validation of DSON files.

	Arguments:
		data (ValidationData): validation data of product to validate.
	"""

	dson: dict[str, Any]

	def _validate(self) -> None:
		"""Perform validation."""

		super()._validate()

		# Cache these
		self.ext_file_parsers = [ ({'0': False, '1': True}[(x := line.split(maxsplit=1))[0]], parse(x[1])) for line in read_list_from('daz/dson_external_file_references.txt') ]
		self.geometry_parser = parse('$.geometry_library[*].id')
		self.uvs_parser = parse('$.uv_set_library[*].id')
		self.morphs_parser = parse('$.modifier_library[*].id')
		self.material_daz_brick_parser = parse('$.material_library[?(@.extra[*].type == "studio/material/daz_brick")].id')
		self.asset_type_parser = parse('$.asset_info.type')
		self.preset_shader_parser = parse('$.scene.materials[*].extra[*].type')
		self.preset_shader_count = parse('$.scene.materials[*].extra[*].`len`')
		self.active_morph_parser = parse('$.modifier_library[?(@.channel.value != 0 & @.channel.value != 0)].channel')
		self.probable_path = re.compile(r'\b(data|runtime)\/', re.IGNORECASE)

		self.invalid_dson_files: dict[str, Exception] = {}
		self.asset_id_mismatch_files: dict[str, str] = {}
		self.geometry_in_duf_files: dict[str, list[str]] = {}
		self.uvs_in_duf_files: dict[str, list[str]] = {}
		self.morphs_in_duf_files: dict[str, list[str]] = {}
		self.materials_in_duf_files: dict[str, list[str]] = {}
		self.active_morphs_in_dsf_files: dict[str, list[tuple[str, str]]] = {}
		for filename in self.data.product_fs.walk.files(filter=['*.dsf', '*.duf']):  # pyright: ignore[reportUnknownMemberType]
			dson: dict[str, Any] | None = None

			if self.data.dson_schema:
				with decompressDSON(self.data.product_fs.openbin(filename)) as file:
					try:
						dson = SchemaCheckedJSON(file, 'dson.schema.json').data
					except Exception as e:
						self.invalid_dson_files[filename] = e

			if not dson:
				with decompressDSON(self.data.product_fs.openbin(filename)) as file:
					try:
						dson = json.load(file)
						if not isinstance(dson, dict):
							raise TypeError('Object expected')
					except Exception as e:
						self.invalid_dson_files[filename] = e
	
			if dson:
				self.dson = dson
				self.asset_type: str = next(iter(self.asset_type_parser.find(dson))).value  # pyright: ignore[reportUnknownArgumentType]

				self._getContributors()
				self._checkAssetID(filename)
				self._checkDSONFileReferences(filename)

				if filename.lower().endswith('.dsf'):
					self._checkActiveMorphsInDSF(filename)

				elif filename.lower().endswith('.duf'):
					self._getShaderTypeInDUF(filename)
					self._checkSupportAssetsInDUF(filename)

		if self.invalid_dson_files:
			self._addIssue(issues.InvalidDSONFilesIssue(self.invalid_dson_files))

		if self.asset_id_mismatch_files:
			self._addIssue(issues.AssetIDMismatchFilesIssue(self.asset_id_mismatch_files))

		if self.geometry_in_duf_files:
			self._addIssue(issues.GeometryInDUFFilesIssue(self.geometry_in_duf_files))

		if self.uvs_in_duf_files:
			self._addIssue(issues.UVSetInDUFFilesIssue(self.uvs_in_duf_files))

		if self.morphs_in_duf_files:
			self._addIssue(issues.MorphInDUFFilesIssue(self.morphs_in_duf_files))

		if self.materials_in_duf_files:
			self._addIssue(issues.ShaderInDUFFilesIssue(self.materials_in_duf_files))

		if self.active_morphs_in_dsf_files:
			self._addIssue(issues.ActiveMorphsInDSFFilesIssue(self.active_morphs_in_dsf_files))

	@rule
	def _getContributors(self) -> None:
		"""Get contributors."""

		try:
			contributor: dict[str, str] = self.dson['asset_info']['contributor']
			author_data = self.data.contributors.setdefault(contributor.get('author', ''), {})
			author_data.setdefault('Email', set()).add(contributor.get('email', ''))
			author_data.setdefault('Website', set()).add(contributor.get('website', ''))
		except Exception:
			pass

	@rule
	def _checkAssetID(self, filename: str) -> None:
		"""Check asset ID."""

		try:
			asset_id = unquote(self.dson['asset_info']['id'])
		except Exception:
			asset_id = ''

		if not filename.lower().endswith(asset_id.lower()):
			self.asset_id_mismatch_files[filename] = asset_id

	@rule
	def _checkDSONFileReferences(self, filename: str) -> None:
		"""Check DSON file references."""

		# TODO: this is slow. Try JMESPath? dpath? multithread? ... something...
		url_references: set[str] = set()
		for encoded, parser in self.ext_file_parsers:
			for v in parser.find(self.dson):
				value = v.value
				if isinstance(value, str):
					if encoded:
						if '://' in value:
							path = unquote(urlparse(value).path.split(':', 1)[-1])
						else:
							path = unquote(urlparse(value.split(':', 1)[-1]).path)

						if ' ' in value:
							warn(f'Possible unquoted attribute marked as quoted: {filename}: {str(v.full_path)}, value: {value}')
						if not path and self.probable_path.search(value):
							warn(f'Possible misparse of path: {filename}: {str(v.full_path)}, value: {value}')
					else:
						path = value

						if '%20' in value:
							warn(f'Possible quoted attribute marked as unquoted: {filename}: {str(v.full_path)}, value: {value}')

					url_references.add(path)

					if str(v.path) in ['AssetFile', 'PresetFile']:
						self.data.postload_files.add(path)

		for file in url_references - {''}:
			if not trackDependencyIfExists(self.data, file, filename):
				self.data.missing_referenced_files.setdefault(file, set()).add(filename)

	@rule
	def _getShaderTypeInDUF(self, filename: str) -> None:
		"""Get shader types in DUF."""

#		if self.asset_type in ['preset_material', 'preset_shader', 'preset_hierarchical_material']:
		shaderCount = sum([v.value for v in self.preset_shader_count.find(self.dson)])
		shaderList: list[str] = [v.value for v in self.preset_shader_parser.find(self.dson)]
		shaders: set[str] = set()

		if shaderCount == 0 or shaderCount < len(shaderList) :
			shaders.add('studio/material/daz_shader')
		shaders.update(shaderList)
		shaders.discard('studio_material_channels')

		if shaders:
			self.data.shader_users[filename] = shaders

	@rule
	def _checkSupportAssetsInDUF(self, filename: str) -> None:
		"""Check DUF file for data better saved in DSF."""

		geometry: list[str] = [v.value for v in self.geometry_parser.find(self.dson)]
		uvs: list[str] = [v.value for v in self.uvs_parser.find(self.dson)]
		morphs: list[str] = [v.value for v in self.morphs_parser.find(self.dson)]
		materials: list[str] = [v.value for v in self.material_daz_brick_parser.find(self.dson)]

		if geometry:
			self.geometry_in_duf_files[filename] = geometry
		if uvs:
			self.uvs_in_duf_files[filename] = uvs
		if morphs:
			self.morphs_in_duf_files[filename] = morphs
		if materials:
			self.materials_in_duf_files[filename] = materials

	@rule
	def _checkActiveMorphsInDSF(self, filename: str) -> None:
		"""Check for active morphs in DSF."""

		for channel in self.active_morph_parser.find(self.dson):
			self.active_morphs_in_dsf_files.setdefault(filename, []).append((channel.value['label'], channel.value['value']))


import re

from lxml import etree
from lxml.etree import DocumentInvalid
#from warnings import warn

from ..issues import metadatafile as issues
from ..validator.resources import read_list_from
from ..validator.ruleset import ProductRuleset, rule
from ..validator.schema import SchemaCheckedXML
from ..validator.utilities import alpha_numeric

_SUPPORT_DIR = 'Runtime/Support'

class ValidateMetadataFiles(ProductRuleset):
	"""Perform metadata file validation of product.

	Arguments:
		data (ValidationData): validation data of product to validate.
	"""

	def _validate(self) -> None:
		"""Perform validation."""

		super()._validate()
		if self.data.product_fs.isdir(_SUPPORT_DIR):
			self.listed_assets: list[str] = []
			self.invalid_metadata_files: dict[str, Exception] = {}
			self.metadata_filenames_mismatches: dict[str, list[str]] = {}
			self.product_name_mismatches: list[str] = []
			self.multiple_stores: dict[str, set[str]] = {}
			self.store_not_daz: dict[str, set[str]] = {}
			self.store_is_daz: list[str] = []
			self.store_not_local: dict[str, set[str]] = {}
			self.missing_product_token: dict[str, list[str]] = {}
			self.missing_artists: dict[str, list[str]] = {}
			self.unexpected_assets: dict[str, list[str]] = {}
			self.unexpected_support_assets: dict[str, list[str]] = {}
			self.bad_support_asset_file_ref: dict[str, list[str]] = {}
			self.missing_content_types: dict[str, list[str]] = {}
			self.deprecated_content_types: dict[str, dict[str, str]] = {}
			self.missing_audiences: dict[str, list[str]] = {}
			self.non_standard_audiences: dict[str, dict[str, str]] = {}
			self.missing_categories: dict[str, list[str]] = {}
			self.duplicate_categories_used: dict[str, dict[str, set[str]]] = {}
			self.parent_categories_used: dict[str, dict[str, list[str]]] = {}
			self.special_categories_used: dict[str, dict[str, list[str]]] = {}
			self.deprecated_categories_used: dict[str, dict[str, list[str]]] = {}
			self.user_categories_used: dict[str, dict[str, list[str]]] = {}
			self.missing_compatibilities: dict[str, list[str]] = {}
			self.missing_compatibility_base: dict[str, list[str]] = {}
			self.missing_tags: dict[str, list[str]] = {}
			self.has_description: dict[str, list[str]] = {}
			self.has_userwords: dict[str, dict[str, list[str]]] = {}
			self.wrong_shader_types: dict[str, list[tuple[str, str, str]]] = {}
			self.nonspecific_shader_types: dict[str, list[tuple[str, str]]] = {}

			for self.metadata_file in self.data.product_fs.filterdir(_SUPPORT_DIR, files=['*.dsx']):
				filename = self.metadata_file.make_path(_SUPPORT_DIR)
				try:
					self.metadata = SchemaCheckedXML(self.data.product_fs.openbin(filename), 'Metadata.xsd').tree
				except DocumentInvalid as e:
					self.invalid_metadata_files[self.metadata_file.name] = e
					# Does not match Schema, but is valid XML, so attempt to use it anyway
					self.metadata = etree.parse(self.data.product_fs.openbin(filename), base_url = filename)
				except Exception as e:
					self.invalid_metadata_files[self.metadata_file.name] = e
					continue

				self._getPackageProductsFromMetadata()
				self._checkMetadataFilename()
				self._checkStoresInMetadata()
				self._checkProductTokenInMetadata()
				self._checkArtistsInMetadata()
				self._checkAssetsInMetadata()
				self._checkAssetsContentType()
				self._checkAssetsAudience()
				self._checkAssetsCategories()
				self._checkAssetsCompatibilities()
				self._checkAssetsCompatibilityBase()
				self._checkAssetsObjectCompatibility()
				self._checkAssetsHaveTags()
				self._checkAssetsHaveDescriptions()
				self._checkAssetsHaveUserwords()

			self._checkPackageProductInMetadata()
			self._checkUnlistedAssets()

			if self.invalid_metadata_files:
				self._addIssue(issues.InvalidProductMetadataIssue(self.invalid_metadata_files))

			if self.metadata_filenames_mismatches:
				self._addIssue(issues.MetadataFilenameIssue(self.metadata_filenames_mismatches))

			if self.product_name_mismatches:
				self._addIssue(issues.PackageProductNotMetadataProductIssue(self.product_name_mismatches))

			if self.multiple_stores:
				self._addIssue(issues.MetadataFileContainsMultipleStoreIDsIssue(self.multiple_stores))

			if self.store_not_daz:
				self._addIssue(issues.MetadataStoreIDIsNotDAZ3DIssue(self.store_not_daz))

			if self.store_is_daz:
				self._addIssue(issues.MetadataStoreIDIsDAZ3DIssue(self.store_is_daz))

			if self.store_not_local:
				self._addIssue(issues.MetadataStoreIDIsNotLocalUserIssue(self.store_not_local))

			if self.missing_product_token:
				self._addIssue(issues.MetadataMissingProductTokenIssue(self.missing_product_token))

			if self.missing_artists:
				self._addIssue(issues.MetadataMissingArtistIssue(self.missing_artists))

			if self.unexpected_assets:
				self._addIssue(issues.MetadataUnexpectedAssetsIssue(self.unexpected_assets))

			if self.unexpected_support_assets:
				self._addIssue(issues.MetadataUnexpectedSupportAssetsIssue(self.unexpected_support_assets))

			if self.bad_support_asset_file_ref:
				self._addIssue(issues.MetadataInvalidSupportAssetFileRefIssue(self.bad_support_asset_file_ref))

			if self.missing_content_types:
				self._addIssue(issues.MetadataMissingContentTypeIssue(self.missing_content_types))

			if self.deprecated_content_types:
				self._addIssue(issues.MetadataDeprecatedContentTypeIssue(self.deprecated_content_types))

			if self.wrong_shader_types:
				self._addIssue(issues.MetadataWrongShaderTypeIssue(self.wrong_shader_types))

			if self.nonspecific_shader_types:
				self._addIssue(issues.MetadataNonspecificShaderTypeIssue(self.nonspecific_shader_types))

			if self.missing_audiences:
				self._addIssue(issues.MetadataMissingAudienceIssue(self.missing_audiences))

			if self.non_standard_audiences:
				self._addIssue(issues.MetadataNonStandardAudienceIssue(self.non_standard_audiences))

			if self.missing_categories:
				self._addIssue(issues.MetadataMissingCategoriesIssue(self.missing_categories))

			if self.duplicate_categories_used:
				self._addIssue(issues.MetadataDuplicateCategoriesUsedIssue(self.duplicate_categories_used))

			if self.parent_categories_used:
				self._addIssue(issues.MetadataParentCategoriesUsedIssue(self.parent_categories_used))

			if self.special_categories_used:
				self._addIssue(issues.MetadataSpecialCategoriesUsedIssue(self.special_categories_used))

			if self.deprecated_categories_used:
				self._addIssue(issues.MetadataDeprecatedCategoriesIssue(self.deprecated_categories_used))

			if self.user_categories_used:
				self._addIssue(issues.MetadataUserCategoriesUsedIssue(self.user_categories_used))

			if self.missing_compatibilities:
				self._addIssue(issues.MetadataMissingCompatibilitiesIssue(self.missing_compatibilities))

			if self.missing_compatibility_base:
				self._addIssue(issues.MetadataMissingCompatibilityBaseIssue(self.missing_compatibility_base))

			if self.missing_tags:
				self._addIssue(issues.MetadataMissingTagsIssue(self.missing_tags))

			if self.has_description:
				self._addIssue(issues.MetadataHasDescriptionIssue(self.has_description))

			if self.has_userwords:
				self._addIssue(issues.MetadataHasUserwordsIssue(self.has_userwords))

	@rule
	def _getPackageProductsFromMetadata(self) -> None:
		"""Get package product names from metadata.

		Updates `data.metadata.products` with found products.
		"""

		self.data.metadata.products.extend([e.attrib['VALUE'] for e in self.metadata.xpath("/ContentDBInstall/Products/Product")])

	@rule
	def _checkMetadataFilename(self) -> None:
		"""Check metadata filename matches name auto generated by Daz Studio."""

		# NOTE: This is a guess and may need to be expanded with additional characters
		REPLACE_METADATA_FILENAME_CHARACTERS = r'[\\/*?:"<>|!. ]'

		expected_metadata_filenames: list[str] = []
		for e in self.metadata.xpath("/ContentDBInstall/Products/Product"):
			storeID = e.xpath('StoreID')[0].attrib['VALUE']
			token = eToken[0].attrib['VALUE'] if (eToken := e.xpath('ProductToken')) else None
			name = e.attrib['VALUE']
			metadata_filename_base = f"{storeID}_{f'{token}_' if token else ''}{name}"
			expected_metadata_filenames.append(re.sub(REPLACE_METADATA_FILENAME_CHARACTERS, '_', metadata_filename_base) + '.dsx')

		if self.metadata_file.name not in expected_metadata_filenames:
			self.metadata_filenames_mismatches[self.metadata_file.name] = expected_metadata_filenames

	@rule
	def _checkStoresInMetadata(self) -> None:
		"""Check stores listed in metadata.

		- Only one unique store listed
		- Store is `DAZ 3D` for products distributed by DAZ
		- Store is NOT `DAZ 3D` for products not distributed by DAZ
		- Store is one of [`DAZ 3D`, `LOCAL USER`, ``]
		This is a NOTICE, not a WARNING as the store will be automatically updated
		to `LOCAL USER` if the store is not already in the users database.

		Updates `data.metadata.stores` with found stores.
		"""

		if len(stores := {e.attrib['VALUE'] for e in self.metadata.xpath("/ContentDBInstall/Products/Product/StoreID")}) > 1:
			self.multiple_stores[self.metadata_file.name] = stores

		self.data.metadata.stores.update(stores)

		if self.data.daz:
			if stores != {'DAZ 3D'}:
				self.store_not_daz[self.metadata_file.name] = stores
		else:
			if 'DAZ 3D' in stores:
				self.store_is_daz.append(self.metadata_file.name)
			if not stores.issubset(defaultStores := {'DAZ 3D', 'LOCAL USER', ''}):
				self.store_not_local[self.metadata_file.name] = stores - defaultStores

	@rule
	def _checkProductTokenInMetadata(self) -> None:
		"""Check product token listed in metadata."""

		if missing_product_tokens := [e.attrib['VALUE'] for e in self.metadata.xpath("/ContentDBInstall/Products/Product[not (StoreID[@VALUE='LOCAL USER']) and not(ProductToken)]")]:
			self.missing_product_token[self.metadata_file.name] = missing_product_tokens

	@rule
	def _checkArtistsInMetadata(self) -> None:
		"""Check artists listed in metadata.

		Updates `data.metadata.artists` with found artists.
		"""

		if missing_artists := [e.attrib['VALUE'] for e in self.metadata.xpath("/ContentDBInstall/Products/Product[not (Artists/Artist)]")]:
			self.missing_artists[self.metadata_file.name] = missing_artists

		self.data.metadata.artists.update({e.attrib['VALUE'] for e in self.metadata.xpath("/ContentDBInstall/Products/Product/Artists/Artist")})

	@rule
	def _checkAssetsInMetadata(self) -> None:
		"""Check assets and support assets listed in metadata."""

		assets = [e.attrib['VALUE'].lstrip('/') for e in self.metadata.xpath("/ContentDBInstall/Products/Product/Assets/Asset")]
		if unexpected_assets := [f for f in assets if not self.data.product_fs.exists(f)]:
			self.unexpected_assets[self.metadata_file.name] = unexpected_assets
		self.listed_assets.extend(assets)

		support_assets = [e.attrib['VALUE'].lstrip('/') for e in self.metadata.xpath("/ContentDBInstall/Products/Product/SupportAssets/SupportAsset")]
		if unexpected_support_assets := [f for f in support_assets if not self.data.product_fs.exists(f)]:
			self.unexpected_support_assets[self.metadata_file.name] = unexpected_support_assets
		self.listed_assets.extend(support_assets)

		support_asset_files = [e.attrib['VALUE'].lstrip('/') for e in self.metadata.xpath("/ContentDBInstall/Products/Product/SupportAssets")]
		if bad_support_asset_file_ref := [f for f in support_asset_files if not self.metadata_file.make_path(_SUPPORT_DIR) == f]:
			self.bad_support_asset_file_ref[self.metadata_file.name] = bad_support_asset_file_ref

	@rule
	def _checkAssetsContentType(self) -> None:
		"""Check asset content type listed in metadata."""

		deprecated_content_type_list = [content_type.lower() for content_type in read_list_from('daz/deprecated_content_types.txt')]
		shader_users_key_lc = {k.lstrip('/').lower(): k for k in self.data.shader_users.keys()}

		# TODO: Expand this type to name map
		shader_map = (('studio/material/daz_shader', 'RSL'), ('studio/material/uber_iray', 'MDL'), ('mt5', 'MT5'), ('mc6', 'MC6'))

		deprecated_content_types: dict[str, str] = {}
		missing_content_types: list[str] = []
		wrong_shader_types: list[tuple[str, str, str]] = []
		nonspecific_shader_types: list[tuple[str, str,]] = []
		for e in self.metadata.xpath("/ContentDBInstall/Products/Product/Assets/Asset"):
			asset = e.attrib['VALUE']
			asset_lc = asset.lower()
			element = e.xpath("ContentType")
			if content_type := element[0].attrib['VALUE'] if len(element) else None:
				if content_type.lower() in deprecated_content_type_list:
					deprecated_content_types[asset] = content_type

				content_type_parts = content_type.split('/')
				if len(content_type_parts) > 1 and content_type_parts[0] == 'Preset' and content_type_parts[1] in ['Materials', 'Shader'] and asset_lc in shader_users_key_lc:
					shaders = self.data.shader_users[shader_users_key_lc[asset_lc]]
#					found_shader = False

					for shader_type, shader_name in shader_map:
						if shaders == {shader_type}:
#							found_shader = True
							if not content_type_parts[-1] == shader_name:
								if len(content_type_parts) == 2 or content_type_parts[-1] == 'Hierarchical':
									nonspecific_shader_types.append((asset, shader_name))
								else:
									wrong_shader_types.append((asset, content_type_parts[-1], shader_name))

#					if not found_shader:
#						warn(f'Unhandled shader: {shaders}')

				# TODO: Validate type better

			else:
				missing_content_types.append(e.attrib['VALUE'])

		if deprecated_content_types:
			self.deprecated_content_types[self.metadata_file.name] = deprecated_content_types

		if missing_content_types:
			self.missing_content_types[self.metadata_file.name] = missing_content_types

		if wrong_shader_types:
			self.wrong_shader_types[self.metadata_file.name] = wrong_shader_types

		if nonspecific_shader_types:
			self.nonspecific_shader_types[self.metadata_file.name] = nonspecific_shader_types

	@rule
	def _checkAssetsAudience(self) -> None:
		"""Check asset audiences listed in metadata."""

		missing_audience: list[str] = []
		non_standard_audience: dict[str, str] = {}
		for e in self.metadata.xpath("/ContentDBInstall/Products/Product/Assets/Asset"):
			element = e.xpath("Audience")
			if audience := element[0].attrib['VALUE'] if len(element) else None:
				if audience not in read_list_from('daz/audiences.txt'):
					non_standard_audience[e.attrib['VALUE']] = audience
			else:
				missing_audience.append(e.attrib['VALUE'])

		if missing_audience:
			self.missing_audiences[self.metadata_file.name] = missing_audience

		if non_standard_audience:
			self.non_standard_audiences[self.metadata_file.name] = non_standard_audience

	@rule
	def _checkAssetsCategories(self) -> None:
		"""Check asset categories listed in metadata."""

		special_categories_list = tuple('/' + category.lower().strip('/') + '/' for category in read_list_from('daz/special_categories.txt'))
		deprecated_categories_list = tuple('/' + category.lower().strip('/') + '/' for category in read_list_from('daz/deprecated_categories.txt'))

		missing_categories: list[str] = []
		duplicate_categories_used: dict[str, set[str]] = {}
		parent_categories_used: dict[str, list[str]] = {}
		special_categories_used: dict[str, list[str]] = {}
		deprecated_categories_used: dict[str, list[str]] = {}
		user_categories_used: dict[str, list[str]] = {}
		for e in self.metadata.xpath("/ContentDBInstall/Products/Product/Assets/Asset"):
			categories: list[str]
			if categories := [element.attrib['VALUE'] for element in e.xpath("Categories/Category")]:
				categories.sort()
				duplicate_categories: set[str] = set()
				parent_categories: list[str] = []
				special_categories: list[str] = []
				user_categories: list[str] = []
				deprecated_categories: list[str] = []
				for i in range(category_count := len(categories)):
					category_lc = '/' + categories[i].lower().strip('/') + '/'

					if (next_index := i + 1) < category_count:
						next_category_lc = '/' + categories[next_index].lower().strip('/') + '/'

						if next_category_lc == category_lc:
							duplicate_categories.add(categories[i])
						elif next_category_lc.startswith(category_lc):
							parent_categories.append(categories[i])

					if category_lc.startswith(special_categories_list):
						special_categories.append(categories[i])

					if category_lc.startswith(deprecated_categories_list):
						deprecated_categories.append(categories[i])

					if not category_lc.startswith('/default/'):
						user_categories.append(categories[i])

					# TODO: Validate categories better

				if duplicate_categories:
					duplicate_categories_used[e.attrib['VALUE']] = duplicate_categories

				if parent_categories:
					parent_categories_used[e.attrib['VALUE']] = parent_categories

				if special_categories:
					special_categories_used[e.attrib['VALUE']] = special_categories

				if deprecated_categories:
					deprecated_categories_used[e.attrib['VALUE']] = deprecated_categories

				if user_categories:
					user_categories_used[e.attrib['VALUE']] = user_categories

			else:
				missing_categories.append(e.attrib['VALUE'])

		if missing_categories:
			self.missing_categories[self.metadata_file.name] = missing_categories

		if duplicate_categories_used:
			self.duplicate_categories_used[self.metadata_file.name] = duplicate_categories_used

		if parent_categories_used:
			self.parent_categories_used[self.metadata_file.name] = parent_categories_used

		if special_categories_used:
			self.special_categories_used[self.metadata_file.name] = special_categories_used

		if deprecated_categories_used:
			self.deprecated_categories_used[self.metadata_file.name] = deprecated_categories_used

		if user_categories_used:
			self.user_categories_used[self.metadata_file.name] = user_categories_used

	@rule
	def _checkAssetsCompatibilities(self) -> None:
		"""Check asset compatibilites listed in metadata."""

		missing_compatibilities: list[str] = []
		for e in self.metadata.xpath("/ContentDBInstall/Products/Product/Assets/Asset"):
			if compatibilities := [element.attrib['VALUE'] for element in e.xpath("Compatibilities/Compatibility")]:
				compatibilities = compatibilities # TODO: Validate compatibilities
			else:
				pass # TODO: Only some types need compatibilities listed
				#missing_compatibilities.append (e.attrib['VALUE'])

		if missing_compatibilities:
			self.missing_compatibilities[self.metadata_file.name] = missing_compatibilities

	@rule
	def _checkAssetsCompatibilityBase(self) -> None:
		"""Check asset compatibility bases listed in metadata."""

		missing_compatibility_base: list[str] = []
		for element in self.metadata.xpath("/ContentDBInstall/Products/Product/Assets/Asset/CompatibilityBase"):
			if compatibility_base := element[0].attrib['VALUE'] if len(element) else None:
				compatibility_base = compatibility_base # TODO: Validate compatibility base
			else:
				pass # TODO: Only some types need compatibility base listed
				#missing_compatibility_base.append (e.attrib['VALUE'])

		if missing_compatibility_base:
			self.missing_compatibility_base[self.metadata_file.name] = missing_compatibility_base

	@rule
	def _checkAssetsObjectCompatibility(self) -> None:
		"""Check asset object compatibility listed in metadata."""

		#missing_object_compatibilities: list[str] = []
		#for e in metadata.xpath("/ContentDBInstall/Products/Product/ObjectCompatibilities/ObjectCompatibility"):
		#	pass # TODO: Validate ObjectCompatibilities

		# TODO: issue if value is #blah and not /blah/blah#blah

	@rule
	def _checkAssetsHaveTags(self) -> None:
		"""Check if assets listed in metadata have tags."""

		missing_tags: list[str] = []
		for e in self.metadata.xpath("/ContentDBInstall/Products/Product/Assets/Asset"):
			if tags := [element.attrib['VALUE'] for element in e.xpath("Tags/Tag")]:
				tags = tags # TODO: Validate tags
			else:
				missing_tags.append(e.attrib['VALUE'])

		if missing_tags:
			self.missing_tags[self.metadata_file.name] = missing_tags

	@rule
	def _checkAssetsHaveDescriptions(self) -> None:
		"""Check if assets listed in metadata have descriptions."""

		if has_description := [e.attrib['VALUE'] for e in self.metadata.xpath("/ContentDBInstall/Products/Product/Assets/Asset[Description[string-length(text()) > 0]]")]:
			self.has_description[self.metadata_file.name] = has_description

	@rule
	def _checkAssetsHaveUserwords(self) -> None:
		"""Check if assets listed in metadata have userwords."""

		for asset in [e.attrib['VALUE'] for e in self.metadata.xpath("/ContentDBInstall/Products/Product/Assets/Asset[Userwords/Userword[string-length(@VALUE) > 0]]")]:
			self.has_userwords.setdefault(self.metadata_file.name, {})[asset] = [e.attrib['VALUE'] for e in self.metadata.xpath(f"/ContentDBInstall/Products/Product/Assets/Asset[@VALUE = '{asset}']/Userwords/Userword[string-length(@VALUE) > 0]")]

	@rule
	def _checkPackageProductInMetadata(self) -> None:
		"""Check package product name is a product listed in metadata."""

		strippedNames = [alpha_numeric(product) for product in self.data.metadata.products]

		for package in self.data.packages:
			if not package.product_name in self.data.metadata.products and not (package.parsed and not package.parsed['name'] in strippedNames):
				self.product_name_mismatches.append(package.product_name or (package.parsed['name'] if package.parsed else ''))

	@rule
	def _checkUnlistedAssets(self) -> None:
		"""Check for unlisted assets."""

		support_dir_lc = _SUPPORT_DIR.lower()
		assets_lc = [asset.lower() for asset in self.listed_assets]

		if unlisted_assets := [file for file in self.data.product_fs.walk.files() if not (f := file.lstrip('/').lower()).startswith(support_dir_lc) and not f in assets_lc]:  # pyright: ignore[reportUnknownMemberType]
			self._addIssue(issues.MetadataUnlistedAssetsIssue(unlisted_assets))


from importlib import resources
from typing import Generator

def read_datafile_bytes(path: str) -> bytes:
	"""Get resource data file from package as bytes."""

	resource = resources.files(__name__.split('.', 1)[0]) / 'data' / path
	return resource.read_bytes()

def read_list_from(path: str) -> Generator[str, None, None]:
	"""Get resource data file from package as text."""

	resource = resources.files(__name__.split('.', 1)[0]) / 'data' / path
	with resource.open() as file:
		for line in file:
			line = line.split(' #')[0].strip()
			if line and not line.startswith('#'):
				yield line


from fs.path import combine, split
from urllib.parse import urlparse

from ..issues import runtime as issues
from ..validator.resources import read_list_from
from ..validator.ruleset import ProductRuleset, rule
from ..validator.utilities import checkDirectoryHasSelfAsChild, checkImageDir, checkVendorDirsOnly

_RUNTIME_DIR = 'Runtime'
_TEXTURES_DIR = 'Runtime/Textures'
_TEMPLATES_DIR = 'Runtime/Templates'
_WEBLINKS_DIR = 'Runtime/WebLinks'

class ValidateRuntimeDirectory(ProductRuleset):
	"""Perform Runtime validation of product.

	Arguments:
		data (ValidationData): validation data of product to validate.
	"""

	def _validate(self) -> None:
		"""Perform validation."""

		super()._validate()
		if self.data.product_fs.isdir(_RUNTIME_DIR):
			self._checkExcessRuntimeDirectory()
			self._checkExcessTexturesDirectory()
			self._checkExcessTemplatesDirectory()
			self._checkUnexpectedDirectoriesInRuntimeDirectory()
			self._checkRootFiles()
			self._checkVendorDirsOnlyInTexturesDirectory()
			self._checkVendorDirsOnlyInTemplatesDirectory()
			self._checkTextureImageFiles()
			self._checkTemplateImageFiles()
			self._checkWebLinks()

	@rule
	def _checkExcessRuntimeDirectory(self) -> None:
		"""Check product for Runtime directory in Runtime directory."""

		if checkDirectoryHasSelfAsChild(self.data.product_fs, _RUNTIME_DIR):
			self._addIssue(issues.ExcessRuntimeDirectoryIssue([combine(_RUNTIME_DIR, split(_RUNTIME_DIR)[1])]))

	@rule
	def _checkExcessTexturesDirectory(self) -> None:
		"""Check product for Textures directory in Runtime/Textures directory."""

		if checkDirectoryHasSelfAsChild(self.data.product_fs, _TEXTURES_DIR):
			self._addIssue(issues.ExcessTexturesDirectoryIssue([combine(_TEXTURES_DIR, split(_TEXTURES_DIR)[1])]))

	@rule
	def _checkExcessTemplatesDirectory(self) -> None:
		"""Check product for Templates directory in Runtime/Templates directory."""

		if checkDirectoryHasSelfAsChild(self.data.product_fs, _TEMPLATES_DIR):
			self._addIssue(issues.ExcessTemplatesDirectoryIssue([combine(_TEMPLATES_DIR, split(_TEMPLATES_DIR)[1])]))

	@rule
	def _checkUnexpectedDirectoriesInRuntimeDirectory(self) -> None:
		"""Check product for unexpected directories in Runtime."""

		expected_runtime_directories = [l.lower() for l in read_list_from('daz/expected_runtime_directories.txt')]
		if self.data.poser:
			expected_runtime_directories.extend([l.lower() for l in read_list_from('poser/expected_runtime_directories.txt')])

		if unexpected_dirs := [entry.name for entry in self.data.product_fs.scandir(_RUNTIME_DIR) if entry.is_dir and entry.name.lower() not in expected_runtime_directories]:
			self._addIssue(issues.UnexpectedDirectoriesInRuntimeDirectoryIssue(unexpected_dirs))

	@rule
	def _checkRootFiles(self) -> None:
		"""Check product for files in root Runtime directory."""

		if root_files := [entry.name for entry in self.data.product_fs.scandir(_RUNTIME_DIR) if entry.is_file]:
			self._addIssue(issues.FilesInRootOfRuntimeDirectoryIssue(root_files))

	@rule
	def _checkVendorDirsOnlyInTexturesDirectory(self) -> None:
		"""Check product has only Vendor directories in Runtime/Textures.

		Updates `data.vendor_paths` with found vendor directories.
		"""

		if texture_root_files := checkVendorDirsOnly(self.data, _TEXTURES_DIR):
			self._addIssue(issues.FilesInRootOfTexturesDirectoryIssue(texture_root_files))

	@rule
	def _checkVendorDirsOnlyInTemplatesDirectory(self) -> None:
		"""Check product has only Vendor directories in Runtime/Templates.

		Updates `data.vendor_paths` with found vendor directories.
		"""

		if templates_root_files := checkVendorDirsOnly(self.data, _TEMPLATES_DIR):
			self._addIssue(issues.FilesInRootOfTemplatesDirectoryIssue(templates_root_files))

	@rule
	def _checkTextureImageFiles(self) -> None:
		"""Check images files in Runtime/Textures."""

		# This is a guess
		PREFERRED_TEXTURE_SUFFIXES = set(['.png', '.exr', '.hdr', '.jpeg', '.jpg', '.tiff', '.tif', '.tga', '.dsi'])

		(non_image_texture_files,
		atypical_image_texture_files,
		unreadable_image_texture_files,
		incorrect_image_texture_file_extensions) = checkImageDir(self.data.product_fs, _TEXTURES_DIR, PREFERRED_TEXTURE_SUFFIXES)

		if non_image_texture_files:
			self._addIssue(issues.NonImageFilesInTexturesDirectoryIssue(non_image_texture_files))

		if atypical_image_texture_files:
			self._addIssue(issues.AtypicalImageFilesInTexturesDirectoryIssue(atypical_image_texture_files))

		if unreadable_image_texture_files:
			self._addIssue(issues.UnreadableImageFilesInTexturesDirectoryIssue(unreadable_image_texture_files))

		if incorrect_image_texture_file_extensions:
			self._addIssue(issues.ImageHasIncorrectFileExtensionIssue(incorrect_image_texture_file_extensions))

		# TODO: check bump and displacement maps are 16+bit greyscale? normal maps 16+bit? ensure none are jpg?
		# Use wand / imagemagick instead of Pillow?

	@rule
	def _checkTemplateImageFiles(self) -> None:
		"""Check images files in Runtime/Templates."""

		# This is a guess
		PREFERRED_TEMPLATE_SUFFIXES = set(['.svg', '.png', '.jpeg', '.jpg', '.tiff', '.tif', '.tga',])

		(non_image_template_files,
		atypical_image_template_files,
		unreadable_image_template_files,
		incorrect_image_template_file_extensions) = checkImageDir(self.data.product_fs, _TEMPLATES_DIR, PREFERRED_TEMPLATE_SUFFIXES)

		if non_image_template_files:
			self._addIssue(issues.NonImageFilesInTemplatesDirectoryIssue(non_image_template_files))

		if atypical_image_template_files:
			self._addIssue(issues.AtypicalImageFilesInTemplatesDirectoryIssue(atypical_image_template_files))

		if unreadable_image_template_files:
			self._addIssue(issues.UnreadableImageFilesInTemplatesDirectoryIssue(unreadable_image_template_files))

		if incorrect_image_template_file_extensions:
			self._addIssue(issues.ImageHasIncorrectFileExtensionIssue(incorrect_image_template_file_extensions))

	@rule
	def _checkWebLinks(self) -> None:
		"""Check product for WebLinks files."""

		if self.data.product_fs.isdir(_WEBLINKS_DIR):
			if self.data.poser:
				if unexpected_files := [file.lstrip('/') for file in self.data.product_fs.walk.files(_WEBLINKS_DIR, exclude=['*.pzs'])]:  # pyright: ignore[reportUnknownMemberType]:
					self._addIssue(issues.UnexpectedFilesInWebLinksDirectoryIssue(unexpected_files))

				invalid_files: list[str] = []
				for file in (file.lstrip('/') for file in self.data.product_fs.walk.files(_WEBLINKS_DIR, filter=['*.pzs'])):  # pyright: ignore[reportUnknownMemberType]:
					try:
						# NOTE: This is just a rough check, it will pass many invalid strings...
						result = urlparse(self.data.product_fs.readtext(file))
						if not all([result.scheme, result.netloc]):
							raise ValueError
					except ValueError:
						invalid_files.append(file)

				if invalid_files:
					self._addIssue(issues.InvalidPZSFileIssue(invalid_files))
			else:
				self._addIssue(issues.WebLinksIssue([_WEBLINKS_DIR]))


import zlib

from ..issues import libraries as issues
from ..validator.resources import read_list_from
from ..validator.ruleset import ProductRuleset, rule
from ..validator.utilities import checkDirectoryHasSelfAsChild

_LIBRARIES_DIR = 'Runtime/Libraries'

class ValidateRuntimeLibrariesDirectory(ProductRuleset):
	"""Perform Runtime/Libraries validation of product.

	Arguments:
		data (ValidationData): validation data of product to validate.
	"""

	def _validate(self) -> None:
		"""Perform validation."""

		super()._validate()
		if self.data.poser and \
		self.data.product_fs.isdir(_LIBRARIES_DIR):
			self._checkExcessLibrariesDirectory()
			self._checkUnexpectedDirectoriesInLibrariesDirectory()
			self._checkRootFiles()
			self._checkValidCompressedFiles()

	@rule
	def _checkExcessLibrariesDirectory(self) -> None:
		"""Check product for Libraries directory in Runtime/Libraries directory."""

		if checkDirectoryHasSelfAsChild(self.data.product_fs, _LIBRARIES_DIR):
			self._addIssue(issues.ExcessRuntimeLibrariesDirectoryIssue())

	@rule
	def _checkUnexpectedDirectoriesInLibrariesDirectory(self) -> None:
		"""Check product for unexpected directories in Runtime/Libraries."""

		expected_runtime_directories = [l.lower() for l in read_list_from('poser/expected_libraries_directories.txt')]

		if unexpected_dirs := [entry.name for entry in self.data.product_fs.scandir(_LIBRARIES_DIR) if entry.is_dir and entry.name.lower() not in expected_runtime_directories]:
			self._addIssue(issues.UnexpectedDirectoriesInRuntimeLibrariesDirectoryIssue(unexpected_dirs))

	@rule
	def _checkRootFiles(self) -> None:
		"""Check product for files in root Runtime/Libraries directory."""

		if root_files := [entry.name for entry in self.data.product_fs.scandir(_LIBRARIES_DIR) if entry.is_file]:
			self._addIssue(issues.FilesInRootOfRuntimeLibrariesDirectoryIssue(root_files))

	@rule
	def _checkValidCompressedFiles(self) -> None:
		"""Check compressed files in Runtime/Libraries directory are valid zlib compressed files."""

		POSER_COMPRESSED_FILE_SUFFIXES = set(['*.cmz', '*.crz', '*.fcz', '*.hdz', '*.hrz', '*.ltz', '*.mcz', '*.mz5', '*.p2z', '*.ppz', '*.pzz'])

		invalid_zlib_files: list[str] = []
		for file in self.data.product_fs.walk.files(_LIBRARIES_DIR, filter=POSER_COMPRESSED_FILE_SUFFIXES):  # pyright: ignore[reportUnknownMemberType]
			try:
				# NOTE: Poser compressed files should be zlib compressed, but DAZ Studio can read gzip as well, so allow it.
				zlib.decompress(self.data.product_fs.readbytes(file), zlib.MAX_WBITS|32)
			except zlib.error:
				invalid_zlib_files.append(file)

		if invalid_zlib_files:
			self._addIssue(issues.InvalidCompressedFilesInRuntimeLibrariesDirectoryIssue(invalid_zlib_files))

	@rule
	def _getContentTypeOfFiles(self) -> None:
		"""Get content type of files."""

		for file in self.data.product_fs.walk.files(_LIBRARIES_DIR, filter=['*.mt5', '*.mz5', '*.mc6', '*.mcz']):  # pyright: ignore[reportUnknownMemberType]
			self.data.shader_users[file] = {'mt5' if file.endswith('5') else 'mc6'}


from . import PackageWarning

class InvalidPackageManifestIssue(PackageWarning):
	title = 'Invalid Package Manifest'
	reference = 'http://docs.daz3d.com/doku.php/public/software/install_manager/referenceguide/tech_articles/package_manifest/'
	description = "Package Manifest 'Manifest.dsx' is invalid"

class MissingPackageManifestFilesIssue(PackageWarning):
	title = 'Missing file(s) in Package'
	reference = 'http://docs.daz3d.com/doku.php/public/software/install_manager/referenceguide/tech_articles/package_manifest/'
	description = "Package Manifest lists file not included in package"

class MissingPackageManifestRecordIssue(PackageWarning):
	title = 'File(s) missing from Package Manifest'
	reference = 'http://docs.daz3d.com/doku.php/public/software/install_manager/referenceguide/tech_articles/package_manifest/'
	description = "Package includes file not listed in Package Manifest"

class RootFilesInPackageIssue(PackageWarning):
	title = 'Unexpected root file(s) in Package'
	description = "All files in Package except for Manifest.dsx and Supplement.dsx must be in subdirectories"
